from typing import ClassVar

from ixmp4 import db
from ixmp4.data import abstract, types
from ixmp4.data.db import mixins

from .. import base, versions


class Region(base.BaseModel, mixins.HasCreationInfo):
    __tablename__ = "region"
    NotFound: ClassVar = abstract.Region.NotFound
    NotUnique: ClassVar = abstract.Region.NotUnique
    DeletionPrevented: ClassVar = abstract.Region.DeletionPrevented

    name: types.UniqueName
    hierarchy: types.String = db.Column(db.String(1023), nullable=False)


class RegionVersion(versions.DefaultVersionModel):
    __tablename__ = "region_version"
    name: types.String = db.Column(db.String(255), nullable=False)
    hierarchy: types.String = db.Column(db.String(1023), nullable=False)

    created_at: types.DateTime = db.Column(nullable=True)
    created_by: types.Username


version_triggers = versions.PostgresVersionTriggers(
    Region.__table__, RegionVersion.__table__
)



# NOTE IndexSet must be imported first s.t. IndexSetVersion is avail in .associations
from ixmp4.data.db.optimization.indexset import IndexSet, IndexSetRepository

from .equation import Equation, EquationRepository
from .parameter import Parameter, ParameterRepository
from .scalar import Scalar, ScalarRepository
from .table import Table, TableRepository
from .variable import Variable, VariableRepository



import logging
from typing import Any, cast

from sqlalchemy import Connection, FromClause, Table, event, schema
from sqlalchemy.orm import Session
from sqlalchemy.sql import ColumnCollection, ColumnElement

from ixmp4.core.exceptions import ProgrammingError

from .ddl import (
    DeleteTrigger,
    InsertTrigger,
    UpdateTrigger,
    VersionProcedure,
)
from .model import DefaultVersionModel, Operation
from .repository import VersionRepository
from .transaction import Transaction, TransactionRepository

logger = logging.getLogger(__name__)
default_tx_id_column_key = "transaction_id"
default_end_tx_id_column_key = "end_transaction_id"
default_op_column_key = "operation_type"


class PostgresVersionTriggers(object):
    version_procedure: VersionProcedure
    insert_trigger: InsertTrigger
    update_trigger: UpdateTrigger
    delete_trigger: DeleteTrigger

    entities: list[DeleteTrigger | InsertTrigger | UpdateTrigger | VersionProcedure]

    def __init__(
        self,
        table: Table | FromClause,
        version_table: Table | FromClause,
        transaction_table: Table = cast(Table, Transaction.__table__),
        transaction_id_column: ColumnElement[int] | None = None,
        end_transaction_id_column: ColumnElement[int] | None = None,
        operation_type_column: ColumnElement[int] | None = None,
    ) -> None:
        if not isinstance(table, Table):
            raise ProgrammingError(
                f"Argument 'table' must be `Table` not {table.__class__.__name__}"
            )
        if not isinstance(version_table, Table):
            raise ProgrammingError(
                "Argument 'version_table' must be `Table` not "
                f"`{version_table.__class__.__name__}`"
            )

        logger.debug(
            "Initializing version triggers for tables "
            f"'{table.name}' and '{version_table.name}'."
        )

        self.table = table
        self.version_table = version_table

        if transaction_id_column is None:
            transaction_id_column = self.column_or_exception(
                "transaction_id_column", default_tx_id_column_key, version_table
            )
        if end_transaction_id_column is None:
            end_transaction_id_column = self.column_or_exception(
                "end_transaction_id_column", default_end_tx_id_column_key, version_table
            )
        if operation_type_column is None:
            operation_type_column = self.column_or_exception(
                "operation_type_column", default_op_column_key, version_table
            )

        self.check_primary_key(table, version_table, transaction_id_column)

        versioned_columns: ColumnCollection[str, ColumnElement[Any]] = (
            ColumnCollection()
        )
        for column in table.columns:
            version_column = version_table.columns.get(column.key)
            if version_column is not None and self.column_corresponds(
                column, version_column
            ):
                versioned_columns.add(column)

        self.version_procedure = VersionProcedure(
            table,
            version_table,
            transaction_table,
            versioned_columns,
            transaction_id_column,
            end_transaction_id_column,
        )
        self.insert_trigger = InsertTrigger(table, self.version_procedure)
        self.update_trigger = UpdateTrigger(table, self.version_procedure)
        self.delete_trigger = DeleteTrigger(table, self.version_procedure)
        self.entities = [
            self.version_procedure,
            self.insert_trigger,
            self.update_trigger,
            self.delete_trigger,
        ]

        self.create_listeners()

    def column_or_exception(
        self, arg_name: str, def_col_name: str, table: Table
    ) -> ColumnElement[Any]:
        try:
            return table.columns[def_col_name]
        except KeyError:
            raise ProgrammingError(
                f"Provide the `{arg_name}` argument or add `{def_col_name}`"
                f"column to version table '{table.name}'.."
            )

    def column_corresponds(
        self, col: ColumnElement[Any], other: ColumnElement[Any]
    ) -> bool:
        return str(col.type) == str(other.type)

    def check_primary_key(
        self,
        table: Table,
        version_table: Table,
        transaction_id_column: ColumnElement[int],
    ) -> None:
        expected_version_pk = ColumnCollection(table.primary_key.columns.items())
        expected_version_pk.add(transaction_id_column)
        current_pk = version_table.primary_key.columns

        for expected_col in expected_version_pk:
            current_col = current_pk.get(expected_col.key)
            if current_col is None or not self.column_corresponds(
                expected_col, current_col
            ):
                raise ProgrammingError(
                    "Version table primary key must consist of "
                    f"original primary key and '{transaction_id_column.key}'.\n"
                    f"Expected: {expected_version_pk}\n"
                    f"Current: {current_pk}"
                )

    def sync_entities(self, con: Session | Connection) -> None:
        self.drop_entities(con)
        self.create_entities(con)

    def create_entities(self, con: Session | Connection) -> None:
        for ent in self.entities:
            ddl = schema.DDL(ent.to_create_sql()).execute_if(dialect="postgresql")  # type: ignore[no-untyped-call]
            con.execute(ddl)

    def drop_entities(self, con: Session | Connection) -> None:
        for ent in reversed(self.entities):
            ddl = schema.DDL(ent.to_drop_sql()).execute_if(  # type: ignore[no-untyped-call]
                dialect="postgresql"
            )
            con.execute(ddl)

    def create_listeners(self) -> None:
        """Creates listeners for trigger creation when using
        `BaseModel.metadata.create_all(...)`, for example when testing."""
        for ent in self.entities:
            ddl = schema.DDL(ent.to_create_sql()).execute_if(  # type: ignore[no-untyped-call]
                dialect="postgresql"
            )
            event.listen(self.table, "after_create", ddl)

        for ent in reversed(self.entities):
            ddl = schema.DDL(ent.to_drop_sql()).execute_if(  # type: ignore[no-untyped-call]
                dialect="postgresql"
            )
            event.listen(self.table, "before_drop", ddl)



from typing import TYPE_CHECKING, ClassVar

from ixmp4 import db
from ixmp4.data import abstract, types
from ixmp4.data.db.model.model import Model
from ixmp4.data.db.scenario.model import Scenario

from .. import base, mixins, versions

if TYPE_CHECKING:
    from ixmp4.data.db.optimization.equation import Equation
    from ixmp4.data.db.optimization.indexset import IndexSet
    from ixmp4.data.db.optimization.parameter import Parameter
    from ixmp4.data.db.optimization.scalar import Scalar
    from ixmp4.data.db.optimization.table import Table
    from ixmp4.data.db.optimization.variable import Variable as OptimizationVariable


class Run(base.BaseModel, mixins.HasUpdateInfo):
    __tablename__ = "run"
    IsLocked: ClassVar = abstract.Run.IsLocked
    NotFound: ClassVar = abstract.Run.NotFound
    NotUnique: ClassVar = abstract.Run.NotUnique
    DeletionPrevented: ClassVar = abstract.Run.DeletionPrevented

    NoDefaultVersion: ClassVar = abstract.Run.NoDefaultVersion

    __table_args__ = (db.UniqueConstraint("model__id", "scenario__id", "version"),)

    model__id: types.Integer = db.Column(
        db.ForeignKey("model.id"), nullable=False, index=True
    )
    model: types.Mapped[Model] = db.relationship(
        "Model",
        backref="run",
        foreign_keys=[model__id],
    )

    scenario__id: types.Integer = db.Column(
        db.ForeignKey("scenario.id"),
        nullable=False,
        index=True,
    )
    scenario: types.Mapped[Scenario] = db.relationship(
        "Scenario",
        backref="run",
        foreign_keys=[scenario__id],
    )

    equations: types.Mapped[list["Equation"]] = db.relationship()
    indexsets: types.Mapped[list["IndexSet"]] = db.relationship()
    parameters: types.Mapped[list["Parameter"]] = db.relationship()
    scalars: types.Mapped[list["Scalar"]] = db.relationship()
    tables: types.Mapped[list["Table"]] = db.relationship()
    variables: types.Mapped[list["OptimizationVariable"]] = db.relationship()

    version: types.Integer = db.Column(db.Integer, nullable=False)
    is_default: types.Boolean = db.Column(db.Boolean, default=False, nullable=False)
    lock_transaction: types.Mapped[int | None] = db.Column(nullable=True, index=True)


class RunVersion(versions.DefaultVersionModel):
    __tablename__ = "run_version"
    model__id: types.Integer = db.Column(nullable=False, index=True)
    scenario__id: types.Integer = db.Column(nullable=False, index=True)
    version: types.Integer = db.Column(nullable=False)
    is_default: types.Boolean = db.Column(nullable=False)
    lock_transaction: types.Integer = db.Column(nullable=True, index=True)

    created_at: types.DateTime = db.Column(nullable=True)
    created_by: types.Username
    updated_at: types.DateTime = db.Column(nullable=True)
    updated_by: types.Username


version_triggers = versions.PostgresVersionTriggers(Run.__table__, RunVersion.__table__)



from typing import ClassVar

from ixmp4.db import Session, filters, utils

from .. import Run, TimeSeries
from ..base import SelectType


class TimeSeriesFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    id: filters.Id | None = filters.Field(None)

    sqla_model: ClassVar[type] = TimeSeries

    def join(self, exc: SelectType, session: Session | None = None) -> SelectType:
        if not utils.is_joined(exc, TimeSeries):
            exc = exc.join(TimeSeries, onclause=TimeSeries.run__id == Run.id)
        return exc



from typing import TYPE_CHECKING

import pandas as pd

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import TypedDict, Unpack

from ixmp4 import db
from ixmp4.data import abstract
from ixmp4.data.abstract.annotations import HasNameFilter
from ixmp4.data.auth.decorators import guard
from ixmp4.data.db import versions
from ixmp4.db.filters import BaseFilter

from .. import base
from .docs import VariableDocsRepository
from .model import Variable, VariableVersion

if TYPE_CHECKING:
    from ixmp4.data.backend.db import SqlAlchemyBackend


class VariableVersionRepository(versions.VersionRepository[VariableVersion]):
    model_class = VariableVersion


class EnumerateKwargs(HasNameFilter, total=False):
    _filter: BaseFilter


class CreateKwargs(TypedDict, total=False):
    name: str


class VariableRepository(
    base.Creator[Variable],
    base.Retriever[Variable],
    base.Enumerator[Variable],
    abstract.VariableRepository,
):
    model_class = Variable
    versions: VariableVersionRepository

    def __init__(self, *args: "SqlAlchemyBackend") -> None:
        super().__init__(*args)
        self.docs = VariableDocsRepository(*args)
        self.versions = VariableVersionRepository(*args)

        from .filter import VariableFilter

        self.filter_class = VariableFilter

    def add(self, name: str) -> Variable:
        variable = Variable(name=name)
        self.session.add(variable)
        return variable

    @guard("view")
    def get(self, name: str) -> Variable:
        exc = db.select(Variable).where(Variable.name == name)
        try:
            return self.session.execute(exc).scalar_one()
        except db.NoResultFound:
            raise Variable.NotFound

    @guard("edit")
    def create(self, *args: str, **kwargs: Unpack[CreateKwargs]) -> Variable:
        return super().create(*args, **kwargs)

    @guard("view")
    def list(self, **kwargs: Unpack[EnumerateKwargs]) -> list[Variable]:
        return super().list(**kwargs)

    @guard("view")
    def tabulate(self, **kwargs: Unpack[EnumerateKwargs]) -> pd.DataFrame:
        return super().tabulate(**kwargs)



from typing import ClassVar

from ixmp4.db import Session, filters, utils

from .. import Measurand, TimeSeries, Variable
from ..base import SelectType


class VariableFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    id: filters.Id | None = filters.Field(None)
    name: filters.String | None = filters.Field(None)

    sqla_model: ClassVar[type] = Variable

    def join(self, exc: SelectType, session: Session | None = None) -> SelectType:
        if not utils.is_joined(exc, Measurand):
            exc = exc.join(Measurand, TimeSeries.measurand)
        exc = exc.join(Variable, Measurand.variable)
        return exc



import enum

from sqlalchemy import BigInteger, SmallInteger

from ixmp4 import db
from ixmp4.data import types

from .. import base


class Operation(int, enum.Enum):
    INSERT = 0
    UPDATE = 1
    DELETE = 2


class DefaultVersionModel(base.BaseModel):
    __abstract__ = True
    id: types.Integer = db.Column(primary_key=True)
    transaction_id: types.Integer = db.Column(
        BigInteger(), primary_key=True, index=True, nullable=False
    )
    operation_type: types.Integer = db.Column(
        SmallInteger(), index=True, nullable=False
    )
    end_transaction_id: types.Integer = db.Column(
        BigInteger(), nullable=True, index=True
    )



import logging
from collections.abc import Iterable
from typing import ClassVar, TypeVar

from sqlalchemy.exc import NoResultFound

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import TypedDict, Unpack

from ixmp4 import db
from ixmp4.data import abstract, types

from ..auth.decorators import guard
from . import base

logger = logging.getLogger(__name__)


class AbstractDocs(base.BaseModel):
    NotFound: ClassVar = abstract.Docs.NotFound
    NotUnique: ClassVar = abstract.Docs.NotUnique
    DeletionPrevented: ClassVar = abstract.Docs.DeletionPrevented

    __abstract__ = True

    description: types.Mapped[str]

    dimension__id: types.Mapped[int]


def docs_model(model: type[base.BaseModel]) -> type[AbstractDocs]:
    dimension__id = db.Column(
        db.Integer, db.ForeignKey(model.id, ondelete="CASCADE"), unique=True
    )

    table_dict = {
        "NotFound": AbstractDocs.NotFound,
        "NotUnique": AbstractDocs.NotUnique,
        "__tablename__": model.__tablename__ + "_docs",
        "description": db.Column(db.Text, nullable=False),
        "dimension__id": dimension__id,
    }

    DocsModel = type(model.__name__ + "Docs", (AbstractDocs,), table_dict)
    return DocsModel


DocsType = TypeVar("DocsType", bound=AbstractDocs)


class ListKwargs(TypedDict, total=False):
    dimension_id: int | None
    dimension_id__in: Iterable[int] | None


class BaseDocsRepository(
    base.Creator[DocsType],
    base.Retriever[DocsType],
    base.Deleter[DocsType],
    base.Enumerator[DocsType],
    abstract.DocsRepository,
):
    dimension_model_class: ClassVar[type[base.BaseModel]]

    def select(
        self,
        *,
        _exc: db.sql.Select[tuple[DocsType]] | None = None,
        dimension_id: int | None = None,
        dimension_id__in: Iterable[int] | None = None,
    ) -> db.sql.Select[tuple[DocsType]]:
        if _exc is None:
            _exc = db.select(self.model_class)

        # TODO Should we enforce that only one of these filters can be active?
        if dimension_id is not None:
            _exc = _exc.where(self.model_class.dimension__id == dimension_id)
        if dimension_id__in is not None:
            if dimension_id and dimension_id not in dimension_id__in:
                logger.warning(
                    "Applying incompatible filters to select for docs: "
                    f"dimension_id '{dimension_id}' is not in dimension_id__in "
                    f"{dimension_id__in}!"
                )
            _exc = _exc.where(self.model_class.dimension__id.in_(dimension_id__in))

        return _exc

    def select_for_count(
        self,
        _exc: db.sql.Select[tuple[int]],
        dimension_id: int | None = None,
        dimension_id__in: Iterable[int] | None = None,
    ) -> db.sql.Select[tuple[int]]:
        if dimension_id is not None:
            _exc = _exc.where(self.model_class.dimension__id == dimension_id)
        if dimension_id__in is not None:
            if dimension_id and dimension_id not in dimension_id__in:
                logger.warning(
                    "Applying incompatible filters to select for docs: "
                    f"dimension_id '{dimension_id}' is not in dimension_id__in "
                    f"{dimension_id__in}!"
                )
            _exc = _exc.where(self.model_class.dimension__id.in_(dimension_id__in))

        return _exc

    def add(self, dimension_id: int, description: str) -> DocsType:
        docs = self.model_class(description=description, dimension__id=dimension_id)
        self.session.add(docs)
        return docs

    @guard("view")
    def get(self, dimension_id: int) -> DocsType:
        exc = self.select(dimension_id=dimension_id)
        try:
            return self.session.execute(exc).scalar_one()
        except NoResultFound:
            raise self.model_class.NotFound

    @guard("edit")
    def set(self, dimension_id: int, description: str) -> DocsType:
        exc = self.select(dimension_id=dimension_id)
        try:
            docs = self.session.execute(exc).scalar_one()
            docs.description = description
            self.session.commit()
            return docs
        except NoResultFound:
            docs = self.create(dimension_id=dimension_id, description=description)
            return docs

    @guard("edit")
    def delete(self, dimension_id: int) -> None:
        exc = db.delete(self.model_class).where(
            self.model_class.dimension__id == dimension_id
        )

        try:
            self.session.execute(
                exc, execution_options={"synchronize_session": "fetch"}
            )
            self.session.commit()
        # This is weird: if there is no region, we might want to return None
        # for a delete request; but then how do we distinguish a
        # typo also returning None?
        except NoResultFound:
            raise self.model_class.NotFound

    @guard("view")
    def list(self, **kwargs: Unpack[ListKwargs]) -> list[DocsType]:
        return super().list(**kwargs)



from typing import ClassVar

from ixmp4.db import Session, filters, utils

from .. import Model, Run
from ..base import SelectType


class ModelFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    id: filters.Id | None = filters.Field(None)
    name: filters.String | None = filters.Field(None)

    sqla_model: ClassVar[type] = Model

    # TODO using general form here as this seems to be callable on non-Model tables
    def join(self, exc: SelectType, session: Session | None = None) -> SelectType:
        if not utils.is_joined(exc, Model):
            exc = exc.join(Model, Run.model)
        return exc



from typing import Any

from sqlalchemy import Table
from sqlalchemy.sql import ColumnCollection, ColumnElement

from .model import Operation


class VersionProcedure(object):
    signature_template = "%(tablename)s_version_procedure()"
    definition_template = """
    returns trigger as $$
        declare tx_id INT;
    begin
        begin
            insert into %(transaction_tablename)s (issued_at)
            values (NOW()) returning id into tx_id;
        end;

        if (TG_OP='DELETE') then
            update %(version_tablename)s set %(end_transaction_id_column)s = tx_id
            where id in (select id from OLD_TABLE)
            and %(end_transaction_id_column)s is NULL;

            insert into %(version_tablename)s (%(versioned_column_names)s,
            %(transaction_id_column)s, operation_type)
            select %(versioned_column_names)s, tx_id, %(op_delete)s
            from OLD_TABLE;
        elsif (TG_OP='UPDATE') then
            update %(version_tablename)s set %(end_transaction_id_column)s = tx_id
            where id in (select id from NEW_TABLE)
            and %(end_transaction_id_column)s is NULL;

            insert into %(version_tablename)s (%(versioned_column_names)s,
            %(transaction_id_column)s, operation_type)
            select %(versioned_column_names)s, tx_id, %(op_update)s
            from NEW_TABLE;
        elsif (TG_OP='INSERT') then
            insert into %(version_tablename)s (%(versioned_column_names)s,
            %(transaction_id_column)s, operation_type)
            select %(versioned_column_names)s, tx_id, %(op_insert)s
            from NEW_TABLE;
        end if;
        return null;
    end
    $$ language plpgsql;
    """
    definition_context: dict[str, Any]
    signature_context: dict[str, Any]
    definition: str
    signature: str

    def __init__(
        self,
        table: Table,
        version_table: Table,
        transaction_table: Table,
        versioned_columns: ColumnCollection[str, ColumnElement[Any]],
        transaction_id_column: ColumnElement[int],
        end_transaction_id_column: ColumnElement[int],
    ):
        tablename = table.name
        version_tablename = version_table.name
        transaction_tablename = transaction_table.name

        versioned_column_names = ", ".join(versioned_columns.keys())

        transaction_id_column = getattr(
            transaction_id_column, "description", transaction_id_column
        )
        end_transaction_id_column = getattr(
            end_transaction_id_column, "description", end_transaction_id_column
        )
        self.definition_context = {
            "transaction_id_column": transaction_id_column,
            "end_transaction_id_column": end_transaction_id_column,
            "tablename": tablename,
            "version_tablename": version_tablename,
            "transaction_tablename": transaction_tablename,
            "versioned_column_names": versioned_column_names,
            "op_delete": str(Operation.DELETE.value),
            "op_update": str(Operation.UPDATE.value),
            "op_insert": str(Operation.INSERT.value),
        }
        self.signature_context = {"tablename": tablename}
        self.signature = self.signature_template % self.signature_context
        self.definition = self.definition_template % self.definition_context

    def to_create_sql(self) -> str:
        return f"CREATE OR REPLACE FUNCTION {self.signature}\n{self.definition}"

    def to_drop_sql(self) -> str:
        return f"DROP FUNCTION IF EXISTS {self.signature};"


class VersionTrigger(object):
    signature_template: str
    definition_template: str
    tablename: str
    definition_context: dict[str, Any]
    signature_context: dict[str, Any]
    definition: str
    signature: str

    def __init__(self, table: Table, procedure: VersionProcedure):
        self.tablename = table.name

        self.definition_context = {
            "procedure_sig": procedure.signature,
            "tablename": self.tablename,
        }
        self.signature_context = {
            "tablename": self.tablename,
        }

        self.definition = self.definition_template % self.definition_context
        self.signature = self.signature_template % self.signature_context

    def to_create_sql(self) -> str:
        return f"CREATE OR REPLACE TRIGGER {self.signature}\n{self.definition}"

    def to_drop_sql(self) -> str:
        return f"DROP TRIGGER {self.signature} ON {self.tablename};"


class InsertTrigger(VersionTrigger):
    signature_template = "%(tablename)s_version_trigger_insert"
    definition_template = """after insert on %(tablename)s
    referencing
        new table as NEW_TABLE
    for each statement EXECUTE function %(procedure_sig)s;"""


class UpdateTrigger(VersionTrigger):
    signature_template = "%(tablename)s_version_trigger_update"
    definition_template = """after update on %(tablename)s
    referencing
        old table as OLD_TABLE
        new table as NEW_TABLE
    for each statement EXECUTE function %(procedure_sig)s;"""


class DeleteTrigger(VersionTrigger):
    signature_template = "%(tablename)s_version_trigger_delete"
    definition_template = """after delete on %(tablename)s
    referencing
    old table as OLD_TABLE
    for each statement EXECUTE function %(procedure_sig)s;"""



from typing import Any

from ixmp4.data.db.docs import BaseDocsRepository, docs_model

from .model import Variable

VariableDocs = docs_model(Variable)


class VariableDocsRepository(BaseDocsRepository[Any]):
    model_class = VariableDocs
    dimension_model_class = Variable



import logging
from concurrent.futures import ThreadPoolExecutor

import httpx

from ixmp4.conf import settings
from ixmp4.conf.auth import BaseAuth, SelfSignedAuth
from ixmp4.conf.manager import ManagerPlatformInfo, PlatformInfo
from ixmp4.core.exceptions import ImproperlyConfigured, UnknownApiError
from ixmp4.data.api import (
    CheckpointRepository,
    DataPointRepository,
    EquationRepository,
    IndexSetRepository,
    ModelRepository,
    OptimizationVariableRepository,
    ParameterRepository,
    RegionRepository,
    RunMetaEntryRepository,
    RunRepository,
    ScalarRepository,
    ScenarioRepository,
    TableRepository,
    TimeSeriesRepository,
    UnitRepository,
    VariableRepository,
)
from ixmp4.server.rest import APIInfo

from .base import Backend

logger = logging.getLogger(__name__)


class RestBackend(Backend):
    client: httpx.Client
    executor: ThreadPoolExecutor
    timeout: httpx.Timeout

    def __init__(
        self,
        info: PlatformInfo,
        auth: BaseAuth | None = None,
        max_concurrent_requests: int = settings.client_max_concurrent_requests,
    ) -> None:
        super().__init__(info)
        logger.debug(f"Connecting to IXMP4 REST API at {info.dsn}.")
        self.executor = ThreadPoolExecutor(max_workers=max_concurrent_requests)
        self.timeout = httpx.Timeout(settings.client_timeout, connect=60.0)
        self.make_client(info.dsn, auth=auth)
        if isinstance(info, ManagerPlatformInfo):
            if info.notice is not None:
                logger.info("Platform notice: >\n" + info.notice)
        self.create_repositories()

    def make_client(self, rest_url: str, auth: BaseAuth | None) -> None:
        auth = self.get_auth(rest_url, auth)

        self.client = httpx.Client(
            base_url=rest_url,
            timeout=self.timeout,
            http2=True,
            auth=auth,
        )

    def get_auth(
        self, rest_url: str, override_auth: BaseAuth | None
    ) -> BaseAuth | None:
        root = httpx.get(rest_url, follow_redirects=True)
        if root.status_code != 200:
            logger.error("Root API response not OK: " + root.text)
            raise UnknownApiError(f"Server response not OK. ({root.status_code})")

        api_info = APIInfo(**root.json())
        logger.info(f"Connected to IXMP4 Platform '{api_info.name}'")

        import ixmp4

        if ixmp4.__version__ != api_info.version:
            logger.warning(
                "IXMP4 Client and Server versions do not match. "
                f"(Client: {ixmp4.__version__}, Server: {api_info.version})"
            )

        logger.debug("Server UTC Time: " + api_info.utcnow.strftime("%c"))
        logger.debug("Server Is Managed: " + str(api_info.is_managed))
        if api_info.manager_url is not None:
            if (
                api_info.manager_url.rstrip("/")
                != str(settings.manager_url).rstrip("/")
                and api_info.is_managed
            ):
                logger.error(f"Server Manager URL: {api_info.manager_url}")
                logger.error(f"Local Manager URL: {settings.manager_url}")
                raise ImproperlyConfigured(
                    "Trying to connect to a managed REST Platform "
                    "with a mismatching Manager URL."
                )

        if override_auth is None:
            if api_info.is_managed:
                return settings.default_auth
            else:
                logger.info(
                    "Connecting to unmanaged server instance, "
                    "falling back to self-signed auth."
                )
                return SelfSignedAuth(settings.secret_hs256)
        else:
            return override_auth

    def create_repositories(self) -> None:
        self.iamc.datapoints = DataPointRepository(self)
        self.iamc.timeseries = TimeSeriesRepository(self)
        self.iamc.variables = VariableRepository(self)
        self.meta = RunMetaEntryRepository(self)
        self.models = ModelRepository(self)
        self.optimization.equations = EquationRepository(self)
        self.optimization.indexsets = IndexSetRepository(self)
        self.optimization.parameters = ParameterRepository(self)
        self.optimization.scalars = ScalarRepository(self)
        self.optimization.tables = TableRepository(self)
        self.optimization.variables = OptimizationVariableRepository(self)
        self.regions = RegionRepository(self)
        self.runs = RunRepository(self)
        self.scenarios = ScenarioRepository(self)
        self.units = UnitRepository(self)
        self.checkpoints = CheckpointRepository(self)



from typing import TYPE_CHECKING, Any, Literal, TypeVar, overload

from sqlalchemy import sql

from ixmp4.data.db.versions.model import DefaultVersionModel

if TYPE_CHECKING:
    from ixmp4.data.db.optimization.equation.model import Equation, EquationVersion
    from ixmp4.data.db.optimization.indexset.model import IndexSet, IndexSetVersion
    from ixmp4.data.db.optimization.parameter.model import Parameter, ParameterVersion
    from ixmp4.data.db.optimization.scalar.model import Scalar, ScalarVersion
    from ixmp4.data.db.optimization.table.model import Table, TableVersion
    from ixmp4.data.db.optimization.variable.model import (
        OptimizationVariable,
        VariableVersion,
    )
    from ixmp4.data.db.unit import Unit, UnitVersion


def where_recorded_after_transaction(
    exc: sql.Select[Any], transaction__id: int, model_class: type[DefaultVersionModel]
) -> sql.Select[Any]:
    return exc.where(model_class.transaction_id > transaction__id)


def where_valid_at_transaction(
    exc: sql.Select[Any], transaction__id: int, model_class: type[DefaultVersionModel]
) -> sql.Select[Any]:
    return exc.where(
        sql.and_(
            model_class.transaction_id <= transaction__id,
            model_class.operation_type != 2,  # versions.model.Operation.DELETE
            sql.or_(
                model_class.end_transaction_id > transaction__id,
                model_class.end_transaction_id == sql.null(),
            ),
        )
    )


def apply_transaction__id(
    exc: sql.Select[Any],
    model_class: type[DefaultVersionModel],
    transaction__id: int | None,
    valid: Literal["at_transaction", "after_transaction"] = "at_transaction",
) -> sql.Select[Any]:
    _exc = exc

    if transaction__id is not None:
        match valid:
            case "at_transaction":
                _exc = where_valid_at_transaction(_exc, transaction__id, model_class)
            case "after_transaction":
                _exc = where_recorded_after_transaction(
                    _exc, transaction__id, model_class
                )

    return _exc


# NOTE TypeVar doesn't like variables (e.g. sets that expand to these lists, so not
# sure how to avoid this repetition)
NamedModelType = TypeVar("NamedModelType", type["Unit"], type["UnitVersion"])
RunLinkedModelType = TypeVar(
    "RunLinkedModelType",
    type["Equation"],
    type["EquationVersion"],
    type["IndexSet"],
    type["IndexSetVersion"],
    type["Parameter"],
    type["ParameterVersion"],
    type["Scalar"],
    type["ScalarVersion"],
    type["Table"],
    type["TableVersion"],
    type["OptimizationVariable"],
    type["VariableVersion"],
)
SelectModelType = TypeVar(
    "SelectModelType",
    type["Unit"],
    type["UnitVersion"],
    type["Equation"],
    type["EquationVersion"],
    type["IndexSet"],
    type["IndexSetVersion"],
    type["Parameter"],
    type["ParameterVersion"],
    type["Scalar"],
    type["ScalarVersion"],
    type["Table"],
    type["TableVersion"],
    type["OptimizationVariable"],
    type["VariableVersion"],
)


@overload
def select_for_id_map(
    model_class: NamedModelType, run__id: None, transaction__id: int | None = None
) -> sql.Select[tuple[int, str]]: ...


@overload
def select_for_id_map(
    model_class: RunLinkedModelType, run__id: int, transaction__id: int | None = None
) -> sql.Select[tuple[int, str]]: ...


def select_for_id_map(
    model_class: SelectModelType,
    run__id: int | None,
    transaction__id: int | None = None,
) -> sql.Select[tuple[int, str]]:
    exc: sql.Select[tuple[int, str]] = sql.Select(model_class.id, model_class.name)

    if run__id is not None:
        # NOTE I think this is on mypy to not recognize that run__id can only be not
        # None when called with a model_class that has run__id
        exc = exc.where(model_class.run__id == run__id)  # type: ignore[attr-defined]

    if issubclass(model_class, DefaultVersionModel):
        exc = apply_transaction__id(
            exc=exc, transaction__id=transaction__id, model_class=model_class
        )

    return exc



from typing import Any

from ixmp4.data.db.docs import BaseDocsRepository, docs_model

from .model import IndexSet

IndexSetDocs = docs_model(IndexSet)


class IndexSetDocsRepository(BaseDocsRepository[Any]):
    model_class = IndexSetDocs
    dimension_model_class = IndexSet



from fastapi import APIRouter, Body, Depends, Query

from ixmp4.core.exceptions import BadRequest
from ixmp4.data import api
from ixmp4.data.backend.db import SqlAlchemyBackend as Backend
from ixmp4.data.db.iamc.datapoint.filter import DataPointFilter
from ixmp4.data.db.iamc.datapoint.model import DataPoint

from .. import deps
from ..base import EnumerationOutput, Pagination
from ..decorators import autodoc

router: APIRouter = APIRouter(
    prefix="/datapoints",
    tags=["iamc", "datapoints"],
)


@autodoc
@router.patch("/", response_model=EnumerationOutput[api.DataPoint])
def query(
    filter: DataPointFilter = Body(DataPointFilter()),
    join_parameters: bool | None = Query(False),
    join_runs: bool = Query(False),
    join_run_id: bool = Query(False),
    table: bool | None = Query(False),
    pagination: Pagination = Depends(),
    backend: Backend = Depends(deps.get_backend),
) -> EnumerationOutput[DataPoint]:
    """This endpoint is used to retrieve and optionally filter data.add()

    Filter parameters are provided as keyword arguments.

    The available filters can be found here:
    :class:`ixmp4.data.db.iamc.datapoint.filter.DataPointFilter`.

    Examples
    --------

    Filter data points for a given model, scenario combination, and a number of years:

    ..  code-block:: json

        {
            "model" : {"name": "model 1"},
            "scenario" : {"name": "scenario 1"},
            "year__in" : [2020, 2025]
        }

    Return all data for a given variable:

    .. code-block:: json

        {
            "variable": {"name": "Final Energy"}
        }


    """
    if any((join_parameters, join_runs, join_run_id)) and not table:
        raise BadRequest(
            "`join_parameters`, `join_runs` or `join_run_id` can only be used "
            "with `table=true`."
        )

    return EnumerationOutput(
        results=backend.iamc.datapoints.paginate(
            _filter=filter,
            limit=pagination.limit,
            offset=pagination.offset,
            table=bool(table),
            join_parameters=join_parameters,
            join_runs=join_runs,
            join_run_id=join_run_id,
        ),
        total=backend.iamc.datapoints.count(
            _filter=filter,
            join_parameters=join_parameters,
            join_runs=join_runs,
            join_run_id=join_run_id,
        ),
        pagination=pagination,
    )


@autodoc
@router.post("/bulk/")
def bulk_upsert(
    df: api.DataFrame,
    backend: Backend = Depends(deps.get_backend),
) -> None:
    # A pandera.DataFrame is a subclass of pd.DataFrame, so this is fine. Mypy likely
    # complains because our decorators change the type hint in some incompatible way.
    # Might be about covariance again.
    backend.iamc.datapoints.bulk_upsert(df.to_pandas())  # type: ignore[arg-type]


@autodoc
@router.patch("/bulk/")
def bulk_delete(
    df: api.DataFrame,
    backend: Backend = Depends(deps.get_backend),
) -> None:
    backend.iamc.datapoints.bulk_delete(df.to_pandas())  # type: ignore[arg-type]



from typing import Any, Literal, TypeVar

import pandas as pd

from ixmp4 import db
from ixmp4.db.utils.revert import apply_transaction__id

from .. import base
from .model import DefaultVersionModel

VModelType = TypeVar("VModelType", bound=DefaultVersionModel)


class VersionRepository(
    base.Retriever[VModelType],
    base.Enumerator[VModelType],
):
    model_class: type[VModelType]

    def select(
        self,
        transaction__id: int | None = None,
        valid: Literal["at_transaction", "after_transaction"] = "at_transaction",
        **kwargs: Any,
    ) -> db.sql.Select[Any]:
        exc = db.select(self.model_class)

        exc = apply_transaction__id(
            exc=exc,
            model_class=self.model_class,
            transaction__id=transaction__id,
            valid=valid,
        )

        exc = db.utils.where_matches_kwargs(exc, model_class=self.model_class, **kwargs)

        return exc.order_by(self.model_class.id.asc())

    def tabulate(
        self,
        *args: Any,
        _raw: bool = False,
        valid: Literal["at_transaction", "after_transaction"] = "at_transaction",
        **kwargs: Any,
    ) -> pd.DataFrame:
        return super().tabulate(*args, _raw=_raw, valid=valid, **kwargs)



from collections.abc import Iterable

# TODO Use `type` instead of TypeAlias when dropping Python 3.11
from typing import ClassVar, TypeAlias, cast

import pandas as pd

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import Unpack

from ixmp4.data import abstract

from . import base


class RunMetaEntry(base.BaseModel):
    NotFound: ClassVar = abstract.RunMetaEntry.NotFound
    NotUnique: ClassVar = abstract.RunMetaEntry.NotUnique
    DeletionPrevented: ClassVar = abstract.RunMetaEntry.DeletionPrevented

    id: int
    run__id: int
    key: str
    dtype: str
    value: abstract.StrictMetaValue
    value_int: int | None
    value_str: str | None
    value_float: float | None
    value_bool: bool | None


# TODO This is tantalizingly close to the run JsonType, but not quite there.
JsonType: TypeAlias = dict[
    str,
    bool
    | float
    | Iterable[float]
    | abstract.annotations.DefaultFilterAlias
    | dict[
        str,
        bool
        | abstract.annotations.IntFilterAlias
        | dict[str, abstract.annotations.DefaultFilterAlias],
    ]
    | None,
]


class RunMetaEntryRepository(
    base.Creator[RunMetaEntry],
    base.Retriever[RunMetaEntry],
    base.Deleter[RunMetaEntry],
    base.Enumerator[RunMetaEntry],
    base.BulkUpserter[RunMetaEntry],
    base.BulkDeleter[RunMetaEntry],
    abstract.RunMetaEntryRepository,
):
    model_class = RunMetaEntry
    prefix = "meta/"

    def create(self, run__id: int, key: str, value: abstract.MetaValue) -> RunMetaEntry:
        return super().create(run__id=run__id, key=key, value=value)

    def get(self, run__id: int, key: str) -> RunMetaEntry:
        return super().get(run_id=run__id, key=key)

    def delete(self, id: int) -> None:
        super().delete(id)

    def enumerate(
        self, **kwargs: Unpack[abstract.meta.EnumerateKwargs]
    ) -> list[RunMetaEntry] | pd.DataFrame:
        return super().enumerate(**kwargs)

    def list(
        self,
        join_run_index: bool | None = None,
        **kwargs: Unpack[abstract.meta.EnumerateKwargs],
    ) -> list[RunMetaEntry]:
        # base functions require dict, but TypedDict just inherits from Mapping
        json = cast(JsonType, kwargs)
        return super()._list(json=json, params={"join_run_index": join_run_index})

    def tabulate(
        self,
        join_run_index: bool | None = None,
        **kwargs: Unpack[abstract.meta.EnumerateKwargs],
    ) -> pd.DataFrame:
        json = cast(JsonType, kwargs)
        return super()._tabulate(json=json, params={"join_run_index": join_run_index})

    def bulk_upsert(self, df: pd.DataFrame) -> None:
        super().bulk_upsert(df)

    def bulk_delete(self, df: pd.DataFrame) -> None:
        super().bulk_delete(df)



import functools
from collections.abc import Callable
from typing import Any, TypeVar

import pandera as pa
from pandera.errors import SchemaError as PanderaSchemaError
from pandera.typing import DataFrame

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import Unpack

from ixmp4.data.abstract.base import BaseRepository

from .exceptions import SchemaError

T = TypeVar("T")


def check_types(
    func: Callable[[BaseRepository, DataFrame[T]], None],
) -> Callable[[BaseRepository, DataFrame[T]], None]:
    checked_func = pa.check_types(func)

    @functools.wraps(func)
    def wrapper(
        *args: Unpack[tuple[BaseRepository, DataFrame[T]]],
        skip_validation: bool = False,
        **kwargs: Any,
    ) -> None:
        if skip_validation:
            return func(*args, **kwargs)
        try:
            return checked_func(*args, **kwargs)
        except PanderaSchemaError as e:
            raise SchemaError(*e.args)

    return wrapper



from fastapi import APIRouter, Body, Depends, Query

from ixmp4.data import api
from ixmp4.data.backend.db import SqlAlchemyBackend as Backend
from ixmp4.data.db.optimization.variable.filter import OptimizationVariableFilter
from ixmp4.data.db.optimization.variable.model import OptimizationVariable

from .. import deps
from ..base import BaseModel, EnumerationOutput, Pagination
from ..decorators import autodoc

router: APIRouter = APIRouter(
    prefix="/variables",
    tags=["optimization", "variables"],
)


class VariableCreateInput(BaseModel):
    name: str
    run_id: int
    constrained_to_indexsets: str | list[str] | None
    column_names: list[str] | None


class DataInput(BaseModel):
    data: dict[str, list[float] | list[int] | list[str]] | None


@autodoc
@router.get("/{id}/", response_model=api.OptimizationVariable)
def get_by_id(
    id: int,
    backend: Backend = Depends(deps.get_backend),
) -> OptimizationVariable:
    return backend.optimization.variables.get_by_id(id)


@autodoc
@router.patch("/", response_model=EnumerationOutput[api.OptimizationVariable])
def query(
    filter: OptimizationVariableFilter = Body(
        OptimizationVariableFilter(id=None, name=None)
    ),
    table: bool = Query(False),
    pagination: Pagination = Depends(),
    backend: Backend = Depends(deps.get_backend),
) -> EnumerationOutput[OptimizationVariable]:
    return EnumerationOutput(
        results=backend.optimization.variables.paginate(
            _filter=filter,
            limit=pagination.limit,
            offset=pagination.offset,
            table=bool(table),
        ),
        total=backend.optimization.variables.count(_filter=filter),
        pagination=pagination,
    )


@autodoc
@router.patch("/{id}/data/")
def add_data(
    id: int,
    data: DataInput,
    backend: Backend = Depends(deps.get_backend),
) -> None:
    backend.optimization.variables.add_data(id=id, **data.model_dump())


@autodoc
@router.delete("/{id}/data/")
def remove_data(
    id: int,
    data: DataInput,
    backend: Backend = Depends(deps.get_backend),
) -> None:
    backend.optimization.variables.remove_data(id=id, **data.model_dump())


@autodoc
@router.post("/", response_model=api.OptimizationVariable)
def create(
    variable: VariableCreateInput,
    backend: Backend = Depends(deps.get_backend),
) -> OptimizationVariable:
    return backend.optimization.variables.create(**variable.model_dump())


@autodoc
@router.delete("/{id}/")
def delete(
    id: int,
    backend: Backend = Depends(deps.get_backend),
) -> None:
    backend.optimization.variables.delete(id=id)



from .equation import Equation, EquationRepository
from .indexset import IndexSet, IndexSetRepository
from .parameter import Parameter, ParameterRepository
from .scalar import Scalar, ScalarRepository
from .table import Table, TableRepository
from .variable import Variable, VariableRepository



from typing import ClassVar

import numpy as np

from ixmp4 import db
from ixmp4.core.exceptions import OptimizationDataValidationError
from ixmp4.data import types
from ixmp4.data.abstract import optimization as abstract
from ixmp4.data.db import versions

from .. import base


class IndexSet(base.BaseModel):
    __tablename__ = "opt_idx"

    NotFound: ClassVar = abstract.IndexSet.NotFound
    NotUnique: ClassVar = abstract.IndexSet.NotUnique
    DataInvalid: ClassVar = OptimizationDataValidationError
    DeletionPrevented: ClassVar = abstract.IndexSet.DeletionPrevented

    _data_type: types.OptimizationDataType

    _data: types.Mapped[list["IndexSetData"]] = db.relationship(
        back_populates="indexset",
        order_by="IndexSetData.id",
        cascade="all, delete",
        passive_deletes=True,
    )

    @property
    def data(self) -> list[float] | list[int] | list[str]:
        return (
            []
            if self._data_type is None
            else np.array([d.value for d in self._data], dtype=self._data_type).tolist()
        )

    @data.setter
    def data(self, value: list[float] | list[int] | list[str]) -> None:
        return None

    updateable_columns = ["_data_type"]

    run__id: types.RunId

    __table_args__ = (db.UniqueConstraint("name", "run__id"),)


class IndexSetData(base.RootBaseModel):
    __tablename__ = "opt_idx_data"

    indexset: types.Mapped["IndexSet"] = db.relationship(back_populates="_data")
    indexset__id: types.IndexSet__Id
    value: types.String = db.Column(db.String, nullable=False)

    __table_args__ = (db.UniqueConstraint("indexset__id", "value"),)


class IndexSetVersion(versions.DefaultVersionModel):
    __tablename__ = "opt_idx_version"

    name: types.String = db.Column(db.String(255), nullable=False)
    run__id: db.MappedColumn[int] = db.Column(db.Integer, nullable=False, index=True)

    _data_type: types.OptimizationDataType

    created_at: types.DateTime = db.Column(nullable=True)
    created_by: types.Username


class IndexSetDataVersion(versions.DefaultVersionModel):
    __tablename__ = "opt_idx_data_version"

    indexset__id: db.MappedColumn[int] = db.Column(
        db.Integer, nullable=False, index=True
    )
    value: types.String = db.Column(db.String, nullable=False)


version_triggers = versions.PostgresVersionTriggers(
    IndexSet.__table__, IndexSetVersion.__table__
)
data_version_triggers = versions.PostgresVersionTriggers(
    IndexSetData.__table__, IndexSetDataVersion.__table__
)



# type: ignore
"""Create postgres versioning triggers for optimization items

Revision ID: 9596778a8150
Revises: e189ffe6a65e
Create Date: 2025-09-22 18:57:11.623283

"""

import logging

import sqlalchemy as sa
from alembic import op

from ixmp4.data.db.versions import PostgresVersionTriggers

logger = logging.getLogger(__name__)

# Revision identifiers, used by Alembic.
revision = "9596778a8150"
down_revision = "e189ffe6a65e"
branch_labels = None
depends_on = None


transaction_table_name = "transaction"
tables = [
    ("opt_sca", "opt_sca_version"),
    ("opt_idx", "opt_idx_version"),
    ("opt_idx_data", "opt_idx_data_version"),
    ("opt_par", "opt_par_version"),
    ("opt_par_idx_association", "opt_par_idx_association_version"),
    ("opt_tab", "opt_tab_version"),
    ("opt_tab_idx_association", "opt_tab_idx_association_version"),
    ("opt_equ", "opt_equ_version"),
    ("opt_equ_idx_association", "opt_equ_idx_association_version"),
    ("opt_var", "opt_var_version"),
    ("opt_var_idx_association", "opt_var_idx_association_version"),
]


def make_versioning_triggers(
    table_name: str, version_table_name: str, transaction_table: sa.Table
):
    conn = op.get_bind()

    version_table = sa.Table(
        version_table_name,
        sa.MetaData(),
        autoload_with=conn,
    )

    data_table = sa.Table(
        table_name,
        sa.MetaData(),
        autoload_with=conn,
    )
    return PostgresVersionTriggers(data_table, version_table, transaction_table)


def upgrade():
    conn = op.get_bind()
    if conn is None:
        logging.warning("Cannot create version triggers without database connection.")
        return

    dialect_name = conn.dialect.name
    if dialect_name != "postgresql":
        logging.info(
            "Skipping creation of version triggers for "
            f"database dialect '{dialect_name}'."
        )
        return

    transaction_table = sa.Table(
        transaction_table_name,
        sa.MetaData(),
        autoload_with=conn,
    )
    for table, version_table in tables:
        triggers = make_versioning_triggers(table, version_table, transaction_table)
        triggers.create_entities(conn)


def downgrade():
    conn = op.get_bind()
    if conn is None:
        logging.warning("Cannot create version triggers without database connection.")
        return

    dialect_name = conn.dialect.name
    if dialect_name != "postgresql":
        logging.info(
            "Skipping dropping of version triggers for "
            f"database dialect '{dialect_name}'."
        )
        return

    transaction_table = sa.Table(
        transaction_table_name,
        sa.MetaData(),
        autoload_with=conn,
    )
    for table, version_table in tables:
        triggers = make_versioning_triggers(table, version_table, transaction_table)
        triggers.drop_entities(conn)



import logging
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, cast

import pandas as pd

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import Unpack

from ixmp4 import db
from ixmp4.core.exceptions import OptimizationItemUsageError
from ixmp4.data import types
from ixmp4.data.abstract import optimization as abstract
from ixmp4.data.auth.decorators import guard
from ixmp4.data.db import versions
from ixmp4.data.db.optimization.associations import (
    BaseIndexSetAssociationRepository,
    BaseIndexSetAssociationReverter,
    BaseIndexSetAssociationVersionRepository,
)
from ixmp4.data.db.unit import Unit

from .. import base
from .docs import ParameterDocsRepository
from .model import (
    Parameter,
    ParameterIndexsetAssociation,
    ParameterIndexsetAssociationVersion,
    ParameterVersion,
)

if TYPE_CHECKING:
    from ixmp4.data.backend.db import SqlAlchemyBackend

logger = logging.getLogger(__name__)


class ParameterVersionRepository(versions.VersionRepository[ParameterVersion]):
    model_class = ParameterVersion


class ParameterIndexSetAssociationVersionRepository(
    BaseIndexSetAssociationVersionRepository[ParameterIndexsetAssociationVersion]
):
    model_class = ParameterIndexsetAssociationVersion
    parent_version = ParameterVersion
    _item_id_column = "parameter__id"


class ParameterIndexSetAssociationRepository(
    BaseIndexSetAssociationRepository[
        ParameterIndexsetAssociation, ParameterIndexsetAssociationVersion
    ]
):
    model_class = ParameterIndexsetAssociation
    versions: ParameterIndexSetAssociationVersionRepository

    def __init__(self, backend: "SqlAlchemyBackend") -> None:
        super().__init__(backend)

        self.versions = ParameterIndexSetAssociationVersionRepository(backend)

        from .filter import OptimizationParameterIndexSetAssociationFilter

        self.filter_class = OptimizationParameterIndexSetAssociationFilter


class ParameterRepository(
    base.Creator[Parameter],
    base.Deleter[Parameter],
    base.Retriever[Parameter],
    base.Enumerator[Parameter],
    BaseIndexSetAssociationReverter[
        ParameterIndexsetAssociation, ParameterIndexsetAssociationVersion
    ],
    abstract.ParameterRepository,
):
    model_class = Parameter

    UsageError = OptimizationItemUsageError

    versions: ParameterVersionRepository

    def __init__(self, *args: "SqlAlchemyBackend") -> None:
        super().__init__(*args)
        self.docs = ParameterDocsRepository(*args)

        from .filter import OptimizationParameterFilter

        self.filter_class = OptimizationParameterFilter

        self.versions = ParameterVersionRepository(*args)

        self._associations = ParameterIndexSetAssociationRepository(*args)

        self._item_id_column = "parameter__id"

    def add(
        self,
        run_id: int,
        name: str,
        constrained_to_indexsets: list[str],
        column_names: list[str] | None = None,
    ) -> Parameter:
        parameter = Parameter(name=name, run__id=run_id)

        indexsets = {
            indexset: self.backend.optimization.indexsets.get(
                run_id=run_id, name=indexset
            )
            for indexset in set(constrained_to_indexsets)
        }
        for i in range(len(constrained_to_indexsets)):
            _ = ParameterIndexsetAssociation(
                parameter=parameter,
                indexset=indexsets[constrained_to_indexsets[i]],
                column_name=column_names[i] if column_names else None,
            )

        parameter.set_creation_info(auth_context=self.backend.auth_context)
        self.session.add(parameter)

        return parameter

    @guard("view")
    def get(self, run_id: int, name: str) -> Parameter:
        exc = db.select(Parameter).where(
            (Parameter.name == name) & (Parameter.run__id == run_id)
        )
        try:
            return self.session.execute(exc).scalar_one()
        except db.NoResultFound:
            raise Parameter.NotFound

    @guard("view")
    def get_by_id(self, id: int) -> Parameter:
        obj = self.session.get(self.model_class, id)

        if obj is None:
            raise Parameter.NotFound(id=id)

        return obj

    @guard("edit")
    def create(
        self,
        run_id: int,
        name: str,
        constrained_to_indexsets: list[str],
        column_names: list[str] | None = None,
    ) -> Parameter:
        if column_names and len(column_names) != len(constrained_to_indexsets):
            raise self.UsageError(
                f"While processing Parameter {name}: \n"
                "`constrained_to_indexsets` and `column_names` not equal in length! "
                "Please provide the same number of entries for both!"
            )

        if column_names and len(column_names) != len(set(column_names)):
            raise self.UsageError(
                f"While processing Parameter {name}: \n"
                "The given `column_names` are not unique!"
            )

        return super().create(
            run_id=run_id,
            name=name,
            constrained_to_indexsets=constrained_to_indexsets,
            column_names=column_names,
        )

    @guard("edit")
    def delete(self, id: int) -> None:
        # Manually ensure associations are deleted first to fix order of version updates
        associations_to_delete = self._associations.tabulate(parameter_id=id)
        self._associations.bulk_delete(associations_to_delete)
        super().delete(id=id)

    @guard("view")
    def list(self, **kwargs: Unpack["base.EnumerateKwargs"]) -> Iterable[Parameter]:
        return super().list(**kwargs)

    @guard("view")
    def tabulate(self, **kwargs: Unpack["base.EnumerateKwargs"]) -> pd.DataFrame:
        return super().tabulate(**kwargs)

    @guard("edit")
    def add_data(self, id: int, data: dict[str, Any] | pd.DataFrame) -> None:
        if isinstance(data, dict):
            try:
                data = pd.DataFrame.from_dict(data=data)
            except ValueError as e:
                raise Parameter.DataInvalid(str(e)) from e

        if data.empty:
            return  # nothing to do

        parameter = self.get_by_id(id=id)

        missing_columns = set(["values", "units"]) - set(data.columns)
        if missing_columns:
            raise OptimizationItemUsageError(
                "Parameter.data must include the column(s): "
                f"{', '.join(missing_columns)}!"
            )

        # Can use a set for now, need full column if we care about order
        for unit_name in set(data["units"]):
            try:
                self.backend.units.get(name=unit_name)
            except Unit.NotFound as e:
                # TODO Add a helpful hint on how to check defined Units
                raise Unit.NotFound(
                    message=f"'{unit_name}' is not defined for this Platform!"
                ) from e

        index_list = parameter.column_names or parameter.indexset_names
        existing_data = pd.DataFrame(parameter.data)
        if not existing_data.empty:
            existing_data.set_index(index_list, inplace=True)

        parameter.data = cast(
            types.JsonDict,
            (
                data.set_index(index_list).combine_first(existing_data).reset_index()
            ).to_dict(orient="list"),
        )

        self.session.commit()

    @guard("edit")
    def remove_data(self, id: int, data: dict[str, Any] | pd.DataFrame) -> None:
        if isinstance(data, dict):
            data = pd.DataFrame.from_dict(data=data)

        if data.empty:
            return

        parameter = self.get_by_id(id=id)
        index_list = parameter.column_names or parameter.indexset_names
        existing_data = pd.DataFrame(parameter.data)
        if not existing_data.empty:
            existing_data.set_index(index_list, inplace=True)

        # This is the only kind of validation we do for removal data
        try:
            data.set_index(index_list, inplace=True)
        except KeyError as e:
            logger.error(
                f"Data to be removed must include {index_list} as keys/columns, but "
                f"{[name for name in data.columns]} were provided."
            )
            raise OptimizationItemUsageError(
                "The data to be removed must specify one or more complete indices to "
                "remove associated units and values!"
            ) from e

        remaining_data = existing_data[~existing_data.index.isin(data.index)]
        if not remaining_data.index.empty:
            remaining_data.reset_index(inplace=True)

        parameter.data = cast(types.JsonDict, remaining_data.to_dict(orient="list"))

        self.session.commit()

    @guard("edit")
    def revert(self, transaction__id: int, run__id: int) -> None:
        super().revert(transaction__id=transaction__id, run__id=run__id)

        # Revert ParameterIndexSetAssociation
        self._revert_indexset_association(
            repo=self, transaction__id=transaction__id, run__id=run__id
        )



import logging
from collections.abc import Generator
from datetime import datetime, timedelta
from typing import Any, cast
from uuid import uuid4

import httpx
import jwt

from ixmp4.core.exceptions import InvalidCredentials, IxmpError

from .user import User, anonymous_user

logger = logging.getLogger(__name__)


class BaseAuth(object):
    # This should never be called
    def __call__(self, *args: Any, **kwargs: Any) -> httpx.Request:
        raise NotImplementedError

    def auth_flow(self, request: httpx.Request) -> Generator[httpx.Request, Any, None]:
        yield self(request)

    def get_user(self) -> User:
        raise NotImplementedError


class SelfSignedAuth(BaseAuth, httpx.Auth):
    """Generates its own JWT with the supplied secret."""

    def __init__(self, secret: str, username: str = "ixmp4"):
        self.secret = secret
        self.user = User(
            id=-1,
            username=username,
            email="ixmp4@iiasa.ac.at",
            is_staff=True,
            is_superuser=True,
            is_verified=True,
            groups=[],
        )
        self.token = self.get_local_jwt()

    def __call__(self, r: httpx.Request) -> httpx.Request:
        try:
            jwt.decode(self.token, self.secret, algorithms=["HS256"])
        except (jwt.InvalidTokenError, jwt.ExpiredSignatureError):
            self.token = self.get_local_jwt()

        r.headers["Authorization"] = "Bearer " + self.token
        return r

    def get_local_jwt(self) -> str:
        self.jti = uuid4().hex
        return jwt.encode(
            {
                "token_type": "access",
                "exp": self.get_expiration_timestamp(),
                "jti": self.jti,
                "sub": "ixmp4",
                "iss": "ixmp4",
                "user": self.user.model_dump(),
            },
            self.secret,
            algorithm="HS256",
        )

    def get_expiration_timestamp(self) -> int:
        return int((datetime.now() + timedelta(minutes=15)).timestamp())

    def get_user(self) -> User:
        self.user.jti = self.jti
        return self.user


class AnonymousAuth(BaseAuth, httpx.Auth):
    def __init__(self) -> None:
        self.user = anonymous_user
        logger.info("Connecting to service anonymously and without credentials.")

    def __call__(self, r: httpx.Request) -> httpx.Request:
        return r

    def get_user(self) -> User:
        return self.user


class ManagerAuth(BaseAuth, httpx.Auth):
    """Uses the SceSe AAC/Management Service to obtain and refresh a token."""

    def __init__(
        self,
        username: str,
        password: str,
        url: str,
    ):
        self.client = httpx.Client(base_url=url, timeout=10.0, http2=True)
        self.username = username
        self.password = password
        self.obtain_jwt()

    def __call__(self, r: httpx.Request) -> httpx.Request:
        try:
            jwt.decode(
                self.access_token,
                options={"verify_signature": False, "verify_exp": True},
            )
        except jwt.ExpiredSignatureError:
            self.refresh_or_reobtain_jwt()

        r.headers["Authorization"] = "Bearer " + self.access_token
        return r

    def obtain_jwt(self) -> None:
        res = self.client.post(
            "/token/obtain/",
            json={
                "username": self.username,
                "password": self.password,
            },
        )
        if res.status_code >= 400:
            if res.status_code == 401:
                raise InvalidCredentials(
                    "Your credentials were rejected by the "
                    "Scenario Services Management System. "
                    "Check if they are correct and your account is active "
                    "or log out with `ixmp4 logout` to use ixmp4 anonymously."
                )
            else:
                raise IxmpError("Unknown API error: " + res.text)

        json = res.json()
        self.access_token = json["access"]
        self.set_user(self.access_token)
        self.refresh_token = json["refresh"]

    def refresh_or_reobtain_jwt(self) -> None:
        try:
            jwt.decode(
                self.refresh_token,
                options={"verify_signature": False, "verify_exp": True},
            )
            self.refresh_jwt()
        except jwt.ExpiredSignatureError:
            self.obtain_jwt()

    def refresh_jwt(self) -> None:
        res = self.client.post(
            "/token/refresh/",
            json={
                "refresh": self.refresh_token,
            },
        )

        if res.status_code >= 400:
            raise IxmpError("Unknown API error: " + res.text)

        self.access_token = res.json()["access"]
        self.set_user(self.access_token)

    def decode_token(self, token: str) -> dict[str, Any]:
        return cast(
            dict[str, Any],
            jwt.decode(
                token,
                options={"verify_signature": False, "verify_exp": False},
            ),
        )

    def set_user(self, token: str) -> None:
        token_dict = self.decode_token(token)
        user_dict = token_dict["user"]
        self.user = User(**user_dict, jti=token_dict.get("jti"))

    def get_user(self) -> User:
        return self.user



from typing import ClassVar

from ixmp4.data.db.optimization.variable.model import VariableIndexsetAssociation
from ixmp4.db import Session, filters, sql

from .. import OptimizationVariable, Run


class OptimizationVariableFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    id: filters.Id | None = filters.Field(None)
    name: filters.String | None = filters.Field(None)
    run__id: filters.Integer | None = filters.Field(None, alias="run_id")

    sqla_model: ClassVar[type] = OptimizationVariable

    def join(
        self,
        exc: sql.Select[tuple[OptimizationVariable]],
        session: Session | None = None,
    ) -> sql.Select[tuple[OptimizationVariable]]:
        exc = exc.join(Run, onclause=OptimizationVariable.run__id == Run.id)
        return exc


class OptimizationVariableIndexSetAssociationFilter(
    filters.BaseFilter, metaclass=filters.FilterMeta
):
    id: filters.Id | None = filters.Field(None)
    variable__id: filters.Integer | None = filters.Field(None, alias="variable_id")

    sqla_model: ClassVar[type] = VariableIndexsetAssociation



from typing import Any

from ixmp4.data.db.docs import BaseDocsRepository, docs_model

from .model import Equation

EquationDocs = docs_model(Equation)


# NOTE Mypy is a static type checker, but we create the class(es) that would need to go
# in BaseDocsRepository[...] dynamically, so I don't know if there's any way to type
# hint them properly. TypeVar, TypeAlias, type(), type[], and NewType all did not work.
class EquationDocsRepository(BaseDocsRepository[Any]):
    model_class = EquationDocs
    dimension_model_class = Equation



from datetime import datetime, timezone
from typing import TYPE_CHECKING

from ixmp4 import db
from ixmp4.data import types

if TYPE_CHECKING:
    from ixmp4.data.auth.context import AuthorizationContext


class HasCreationInfo:
    """Mixin for adding creation audit info to a model.
    The added fields `created_at`, `created_by` are automatically set
    by ixmp4's sqlalchemy event handlers for 'before_insert' and 'do_orm_execute'.
    """

    created_at: types.DateTime = db.Column(db.DateTime, nullable=True)
    created_by: types.Username

    @staticmethod
    def get_username(auth_context: "AuthorizationContext | None") -> str:
        return auth_context.user.username if auth_context is not None else "@unknown"

    @staticmethod
    def get_timestamp() -> datetime:
        return datetime.now(tz=timezone.utc)

    def set_creation_info(self, auth_context: "AuthorizationContext | None") -> None:
        self.created_at = self.get_timestamp()
        self.created_by = self.get_username(auth_context)


class HasUpdateInfo(HasCreationInfo):
    """Mixin for adding update and creation audit info to a model.
    The added fields `updated_at`, `updated_by` are automatically set
    by ixmp4's sqlalchemy event handlers for 'before_update' and 'do_orm_execute'
    (in addition to the inherited fields from `HasCreationInfo`).
    """

    updated_at: types.DateTime = db.Column(db.DateTime, nullable=True)
    updated_by: types.Username

    def set_update_info(self, auth_context: "AuthorizationContext | None") -> None:
        self.updated_at = self.get_timestamp()
        self.updated_by = self.get_username(auth_context)



# type: ignore
"""TEMPORARY Create all missing optimization items for testing

Revision ID: 0d73f7467dab
Revises: c29289ced488
Create Date: 2024-07-08 14:09:49.174145

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# Revision identifiers, used by Alembic.
revision = "0d73f7467dab"
down_revision = "c29289ced488"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "optimization_equation",
        sa.Column("run__id", sa.Integer(), nullable=False),
        sa.Column(
            "data",
            sa.JSON().with_variant(
                postgresql.JSONB(astext_type=sa.Text()), "postgresql"
            ),
            nullable=False,
        ),
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("created_by", sa.String(length=255), nullable=True),
        sa.ForeignKeyConstraint(
            ["run__id"], ["run.id"], name=op.f("fk_optimization_equation_run__id_run")
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_optimization_equation")),
        sa.UniqueConstraint(
            "name", "run__id", name=op.f("uq_optimization_equation_name_run__id")
        ),
    )
    with op.batch_alter_table("optimization_equation", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_optimization_equation_run__id"), ["run__id"], unique=False
        )

    op.create_table(
        "optimization_optimizationvariable",
        sa.Column("run__id", sa.Integer(), nullable=False),
        sa.Column(
            "data",
            sa.JSON().with_variant(
                postgresql.JSONB(astext_type=sa.Text()), "postgresql"
            ),
            nullable=False,
        ),
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("created_by", sa.String(length=255), nullable=True),
        sa.ForeignKeyConstraint(
            ["run__id"],
            ["run.id"],
            name=op.f("fk_optimization_optimizationvariable_run__id_run"),
        ),
        sa.PrimaryKeyConstraint(
            "id", name=op.f("pk_optimization_optimizationvariable")
        ),
        sa.UniqueConstraint(
            "name",
            "run__id",
            name=op.f("uq_optimization_optimizationvariable_name_run__id"),
        ),
    )
    with op.batch_alter_table(
        "optimization_optimizationvariable", schema=None
    ) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_optimization_optimizationvariable_run__id"),
            ["run__id"],
            unique=False,
        )

    op.create_table(
        "optimization_parameter",
        sa.Column("run__id", sa.Integer(), nullable=False),
        sa.Column(
            "data",
            sa.JSON().with_variant(
                postgresql.JSONB(astext_type=sa.Text()), "postgresql"
            ),
            nullable=False,
        ),
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("created_by", sa.String(length=255), nullable=True),
        sa.ForeignKeyConstraint(
            ["run__id"], ["run.id"], name=op.f("fk_optimization_parameter_run__id_run")
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_optimization_parameter")),
        sa.UniqueConstraint(
            "name", "run__id", name=op.f("uq_optimization_parameter_name_run__id")
        ),
    )
    with op.batch_alter_table("optimization_parameter", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_optimization_parameter_run__id"), ["run__id"], unique=False
        )

    op.create_table(
        "optimization_equation_docs",
        sa.Column("description", sa.Text(), nullable=False),
        sa.Column("dimension__id", sa.Integer(), nullable=True),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["dimension__id"],
            ["optimization_equation.id"],
            name=op.f("fk_optimization_equation_docs_dimension__id_optimizatio_bb11"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_optimization_equation_docs")),
        sa.UniqueConstraint(
            "dimension__id", name=op.f("uq_optimization_equation_docs_dimension__id")
        ),
    )
    op.create_table(
        "optimization_optimizationvariable_docs",
        sa.Column("description", sa.Text(), nullable=False),
        sa.Column("dimension__id", sa.Integer(), nullable=True),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["dimension__id"],
            ["optimization_optimizationvariable.id"],
            name=op.f("fk_optimization_optimizationvariable_docs_dimension__id_ba30"),
        ),
        sa.PrimaryKeyConstraint(
            "id", name=op.f("pk_optimization_optimizationvariable_docs")
        ),
        sa.UniqueConstraint(
            "dimension__id",
            name=op.f("uq_optimization_optimizationvariable_docs_dimension__id"),
        ),
    )
    op.create_table(
        "optimization_parameter_docs",
        sa.Column("description", sa.Text(), nullable=False),
        sa.Column("dimension__id", sa.Integer(), nullable=True),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["dimension__id"],
            ["optimization_parameter.id"],
            name=op.f("fk_optimization_parameter_docs_dimension__id_optimizati_898e"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_optimization_parameter_docs")),
        sa.UniqueConstraint(
            "dimension__id", name=op.f("uq_optimization_parameter_docs_dimension__id")
        ),
    )
    with op.batch_alter_table("optimization_column", schema=None) as batch_op:
        batch_op.add_column(sa.Column("equation__id", sa.Integer(), nullable=True))
        batch_op.add_column(sa.Column("parameter__id", sa.Integer(), nullable=True))
        batch_op.add_column(sa.Column("variable__id", sa.Integer(), nullable=True))
        batch_op.alter_column("table__id", existing_type=sa.INTEGER(), nullable=True)
        batch_op.drop_index("ix_optimization_column_table__id")
        batch_op.create_foreign_key(
            batch_op.f("fk_optimization_column_variable__id_optimization_optimi_2f05"),
            "optimization_optimizationvariable",
            ["variable__id"],
            ["id"],
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_optimization_column_parameter__id_optimization_parameter"),
            "optimization_parameter",
            ["parameter__id"],
            ["id"],
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_optimization_column_equation__id_optimization_equation"),
            "optimization_equation",
            ["equation__id"],
            ["id"],
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("optimization_column", schema=None) as batch_op:
        batch_op.drop_constraint(
            batch_op.f("fk_optimization_column_equation__id_optimization_equation"),
            type_="foreignkey",
        )
        batch_op.drop_constraint(
            batch_op.f("fk_optimization_column_parameter__id_optimization_parameter"),
            type_="foreignkey",
        )
        batch_op.drop_constraint(
            batch_op.f("fk_optimization_column_variable__id_optimization_optimi_2f05"),
            type_="foreignkey",
        )
        batch_op.create_index(
            "ix_optimization_column_table__id", ["table__id"], unique=False
        )
        batch_op.alter_column("table__id", existing_type=sa.INTEGER(), nullable=False)
        batch_op.drop_column("variable__id")
        batch_op.drop_column("parameter__id")
        batch_op.drop_column("equation__id")

    op.drop_table("optimization_parameter_docs")
    op.drop_table("optimization_optimizationvariable_docs")
    op.drop_table("optimization_equation_docs")
    with op.batch_alter_table("optimization_parameter", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_optimization_parameter_run__id"))

    op.drop_table("optimization_parameter")
    with op.batch_alter_table(
        "optimization_optimizationvariable", schema=None
    ) as batch_op:
        batch_op.drop_index(batch_op.f("ix_optimization_optimizationvariable_run__id"))

    op.drop_table("optimization_optimizationvariable")
    with op.batch_alter_table("optimization_equation", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_optimization_equation_run__id"))

    op.drop_table("optimization_equation")
    # ### end Alembic commands ###



from ixmp4.data.db import filters as base
from ixmp4.data.db.run import Run
from ixmp4.db import Session, filters, sql, utils

from .model import OptimizationVariable as Variable
from .model import VariableIndexsetAssociation


class RunFilter(base.RunFilter, metaclass=filters.FilterMeta):
    def join(
        self, exc: sql.Select[tuple[Variable]], session: Session | None = None
    ) -> sql.Select[tuple[Variable]]:
        if not utils.is_joined(exc, Run):
            exc = exc.join(Run, onclause=Variable.run__id == Run.id)
        return exc


class OptimizationVariableFilter(
    base.OptimizationVariableFilter, metaclass=filters.FilterMeta
):
    def join(
        self, exc: sql.Select[tuple[Variable]], session: Session | None = None
    ) -> sql.Select[tuple[Variable]]:
        return exc


class OptimizationVariableIndexSetAssociationFilter(
    base.OptimizationVariableIndexSetAssociationFilter, metaclass=filters.FilterMeta
):
    def join(
        self,
        exc: sql.Select[tuple[VariableIndexsetAssociation]],
        session: Session | None = None,
    ) -> sql.Select[tuple[VariableIndexsetAssociation]]:
        return exc



from collections.abc import Iterable
from typing import Any

from .. import base
from ..docs import AbstractDocs, BaseDocsRepository, docs_model
from .model import Model

ModelDocs = docs_model(Model)


class ModelDocsRepository(BaseDocsRepository[Any], base.BaseRepository[Model]):
    model_class = ModelDocs
    dimension_model_class = Model

    def list(
        self,
        *,
        dimension_id: int | None = None,
        dimension_id__in: Iterable[int] | None = None,
    ) -> list[AbstractDocs]:
        return super().list(
            dimension_id=dimension_id, dimension_id__in=dimension_id__in
        )



from typing import TYPE_CHECKING

import pandas as pd

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import TypedDict, Unpack

from .. import versions

if TYPE_CHECKING:
    from ixmp4.data.backend.db import SqlAlchemyBackend

from ixmp4 import db
from ixmp4.data import abstract
from ixmp4.data.auth.decorators import guard
from ixmp4.data.db.model.model import Model
from ixmp4.db.filters import BaseFilter

from .. import base
from .docs import ScenarioDocsRepository
from .model import Scenario, ScenarioVersion


class ScenarioVersionRepository(versions.VersionRepository[ScenarioVersion]):
    model_class = ScenarioVersion


class EnumerateKwargs(abstract.annotations.HasNameFilter, total=False):
    _filter: BaseFilter


class CreateKwargs(TypedDict, total=False):
    name: str


class ScenarioRepository(
    base.Creator[Scenario],
    base.Retriever[Scenario],
    base.Enumerator[Scenario],
    abstract.ScenarioRepository,
):
    docs: ScenarioDocsRepository
    versions: ScenarioVersionRepository

    model_class = Scenario

    def __init__(self, *args: "SqlAlchemyBackend") -> None:
        super().__init__(*args)
        self.docs = ScenarioDocsRepository(*args)
        self.versions = ScenarioVersionRepository(*args)

        from .filter import ScenarioFilter

        self.filter_class = ScenarioFilter

    def join_auth(
        self, exc: db.sql.Select[tuple[Scenario]]
    ) -> db.sql.Select[tuple[Scenario]]:
        from ixmp4.data.db.run.model import Run

        if not db.utils.is_joined(exc, Run):
            exc = exc.join(Run, onclause=Run.scenario__id == Scenario.id)
        if not db.utils.is_joined(exc, Model):
            exc = exc.join(Model, Run.model)

        return exc

    def add(self, name: str) -> Scenario:
        scenario = Scenario(name=name)
        self.session.add(scenario)
        return scenario

    @guard("view")
    def get(self, name: str) -> Scenario:
        exc = db.select(Scenario).where(Scenario.name == name)
        try:
            return self.session.execute(exc).scalar_one()
        except db.NoResultFound:
            raise Scenario.NotFound

    @guard("edit")
    def create(self, *args: str, **kwargs: Unpack[CreateKwargs]) -> Scenario:
        return super().create(*args, **kwargs)

    @guard("view")
    def list(self, **kwargs: Unpack[EnumerateKwargs]) -> list[Scenario]:
        return super().list(**kwargs)

    @guard("view")
    def tabulate(self, **kwargs: Unpack[EnumerateKwargs]) -> pd.DataFrame:
        return super().tabulate(**kwargs)



from typing import TYPE_CHECKING

import pandas as pd
from sqlalchemy.exc import NoResultFound

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import TypedDict, Unpack

from .. import versions

if TYPE_CHECKING:
    from ixmp4.data.backend.db import SqlAlchemyBackend

from ixmp4.data import abstract
from ixmp4.data.auth.decorators import guard
from ixmp4.db.filters import BaseFilter

from .. import base
from .docs import RegionDocsRepository
from .model import Region, RegionVersion


class RegionVersionRepository(versions.VersionRepository[RegionVersion]):
    model_class = RegionVersion


class EnumerateKwargs(
    abstract.annotations.HasNameFilter,
    abstract.annotations.HasHierarchyFilter,
    total=False,
):
    _filter: BaseFilter


class CreateKwargs(TypedDict, total=False):
    name: str
    hierarchy: str


class RegionRepository(
    base.Creator[Region],
    base.Deleter[Region],
    base.Retriever[Region],
    base.Enumerator[Region],
    abstract.RegionRepository,
):
    model_class = Region
    docs: RegionDocsRepository
    versions: RegionVersionRepository

    def __init__(self, *args: "SqlAlchemyBackend") -> None:
        super().__init__(*args)
        self.docs = RegionDocsRepository(*args)
        self.versions = RegionVersionRepository(*args)

        from .filter import RegionFilter

        self.filter_class = RegionFilter

    def add(self, name: str, hierarchy: str) -> Region:
        region = Region(name=name, hierarchy=hierarchy)
        self.session.add(region)
        return region

    @guard("manage")
    def create(self, *args: str, **kwargs: Unpack[CreateKwargs]) -> Region:
        return super().create(*args, **kwargs)

    @guard("manage")
    def delete(self, *args: int) -> None:
        super().delete(*args)

    @guard("view")
    def get(self, name: str) -> Region:
        exc = self.select().where(Region.name == name)
        try:
            return self.session.execute(exc).scalar_one()
        except NoResultFound:
            raise Region.NotFound

    @guard("view")
    def list(self, **kwargs: Unpack[EnumerateKwargs]) -> list[Region]:
        return super().list(**kwargs)

    @guard("view")
    def tabulate(self, **kwargs: Unpack[EnumerateKwargs]) -> pd.DataFrame:
        return super().tabulate(**kwargs)



from typing import Any

from sqlalchemy import DateTime

from ixmp4 import db
from ixmp4.core.exceptions import NotFound
from ixmp4.data import types

from .. import base


class Transaction(base.BaseModel):
    NotFound = NotFound
    __tablename__ = "transaction"

    issued_at: types.DateTime = db.Column(DateTime())


class TransactionRepository(
    base.Retriever[Transaction],
    base.Enumerator[Transaction],
):
    model_class = Transaction

    def select(self, **kwargs: Any) -> db.sql.Select[tuple[Transaction]]:
        return db.select(self.model_class).order_by(self.model_class.id.desc())

    def latest(self) -> Transaction:
        tx = self.session.execute(self.select().limit(1)).scalars().first()
        if tx is None:
            raise Transaction.NotFound("No transactions in database.")
        return tx



from datetime import datetime
from typing import Protocol

import pandas as pd

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import Unpack

from . import base
from .annotations import (
    HasHierarchyFilter,
    HasIdFilter,
    HasNameFilter,
    IamcRegionFilter,
)
from .docs import DocsRepository


class Region(base.BaseModel, Protocol):
    """Region data model."""

    name: str
    "Unique name of the region."
    hierarchy: str
    "Region hierarchy."

    created_at: datetime
    "Creation date/time."
    created_by: str
    "Creator."

    def __str__(self) -> str:
        return f"<Region {self.id} name={self.name}>"


class EnumerateKwargs(HasHierarchyFilter, HasIdFilter, HasNameFilter, total=False):
    iamc: IamcRegionFilter | bool | None


class RegionRepository(
    base.Creator,
    base.Deleter,
    base.Retriever,
    base.Enumerator,
    Protocol,
):
    docs: DocsRepository

    def create(self, name: str, hierarchy: str) -> Region:
        """Creates a region.

        Parameters
        ----------
        name : str
            The name of the region.
        hierarchy : str
            The hierarchy this region is assigned to.

        Raises
        ------
        :class:`ixmp4.data.abstract.Region.NotUnique`:
            If the region with `name` is not unique.


        Returns
        -------
        :class:`ixmp4.ata.base.iamc.Region`:
            The created region.
        """
        ...

    def delete(self, id: int) -> None:
        """Deletes a region.

        Parameters
        ----------
        id : int
            The unique integer id of the region.

        Raises
        ------
        :class:`ixmp4.data.abstract.Region.NotFound`:
            If the region with `id` does not exist.
        :class:`ixmp4.data.abstract.Region.DeletionPrevented`:
            If the region with `id` is used in the database, preventing it's deletion.
        """
        ...

    def get(self, name: str) -> Region:
        """Retrieves a region.

        Parameters
        ----------
        name : str
            The unique name of the region.

        Raises
        ------
        :class:`ixmp4.data.abstract.Region.NotFound`:
            If the region with `name` does not exist.

        Returns
        -------
        :class:`ixmp4.data.base.iamc.Region`:
            The retrieved region.
        """
        ...

    def get_or_create(self, name: str, hierarchy: str | None = None) -> Region:
        try:
            region = self.get(name)
        except Region.NotFound:
            if hierarchy is None:
                raise TypeError(
                    "Argument `hierarchy` is required if `Region` with `name` does not "
                    "exist."
                )
            return self.create(name, hierarchy)

        if hierarchy is not None and region.hierarchy != hierarchy:
            raise Region.NotUnique(name)
        else:
            return region

    def list(self, **kwargs: Unpack[EnumerateKwargs]) -> list[Region]:
        r"""Lists regions by specified criteria.

        Parameters
        ----------
        \*\*kwargs: any
            Any filter parameters as specified in
            `ixmp4.data.db.region.filter.RegionFilter`.

        Returns
        -------
        Iterable[:class:`ixmp4.data.abstract.Region`]:
            List of regions.
        """
        ...

    def tabulate(self, **kwargs: Unpack[EnumerateKwargs]) -> pd.DataFrame:
        r"""Tabulate regions by specified criteria.

        Parameters
        ----------
        \*\*kwargs: any
            Any filter parameters as specified in
            `ixmp4.data.db.region.filter.RegionFilter`.

        Returns
        -------
        :class:`pandas.DataFrame`:
            A data frame with the columns:
                - id
                - name
        """
        ...



# type: ignore
"""alter version table column nullability

Revision ID: ad25c5b542c4
Revises: 84b534bbc858
"""

import sqlalchemy as sa
from alembic import op

revision = "ad25c5b542c4"
down_revision = "84b534bbc858"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("checkpoint", schema=None) as batch_op:
        batch_op.alter_column(
            "transaction__id", existing_type=sa.INTEGER(), nullable=True
        )

    with op.batch_alter_table(
        "iamc_datapoint_universal_version", schema=None
    ) as batch_op:
        batch_op.alter_column(
            "time_series__id", existing_type=sa.INTEGER(), nullable=False
        )
        batch_op.alter_column(
            "type", existing_type=sa.VARCHAR(length=255), nullable=False
        )

    with op.batch_alter_table("iamc_measurand_version", schema=None) as batch_op:
        batch_op.alter_column(
            "variable__id", existing_type=sa.INTEGER(), nullable=False
        )
        batch_op.alter_column("unit__id", existing_type=sa.INTEGER(), nullable=False)

    with op.batch_alter_table("iamc_timeseries_version", schema=None) as batch_op:
        batch_op.alter_column("region__id", existing_type=sa.INTEGER(), nullable=False)
        batch_op.alter_column(
            "measurand__id", existing_type=sa.INTEGER(), nullable=False
        )
        batch_op.alter_column("run__id", existing_type=sa.INTEGER(), nullable=False)

    with op.batch_alter_table("iamc_variable_version", schema=None) as batch_op:
        batch_op.alter_column(
            "name", existing_type=sa.VARCHAR(length=255), nullable=False
        )

    with op.batch_alter_table("model_version", schema=None) as batch_op:
        batch_op.alter_column(
            "name", existing_type=sa.VARCHAR(length=255), nullable=False
        )

    with op.batch_alter_table("region_version", schema=None) as batch_op:
        batch_op.alter_column(
            "name", existing_type=sa.VARCHAR(length=255), nullable=False
        )
        batch_op.alter_column(
            "hierarchy", existing_type=sa.VARCHAR(length=1023), nullable=False
        )

    with op.batch_alter_table("run_version", schema=None) as batch_op:
        batch_op.alter_column("model__id", existing_type=sa.INTEGER(), nullable=False)
        batch_op.alter_column(
            "scenario__id", existing_type=sa.INTEGER(), nullable=False
        )
        batch_op.alter_column("version", existing_type=sa.INTEGER(), nullable=False)
        batch_op.alter_column("is_default", existing_type=sa.BOOLEAN(), nullable=False)

    with op.batch_alter_table("runmetaentry_version", schema=None) as batch_op:
        batch_op.alter_column("run__id", existing_type=sa.INTEGER(), nullable=False)
        batch_op.alter_column(
            "key", existing_type=sa.VARCHAR(length=1023), nullable=False
        )
        batch_op.alter_column(
            "dtype", existing_type=sa.VARCHAR(length=20), nullable=False
        )

    with op.batch_alter_table("scenario_version", schema=None) as batch_op:
        batch_op.alter_column(
            "name", existing_type=sa.VARCHAR(length=255), nullable=False
        )

    with op.batch_alter_table("transaction", schema=None) as batch_op:
        batch_op.alter_column("issued_at", existing_type=sa.DateTime(), nullable=False)
        batch_op.drop_column("remote_addr")

    with op.batch_alter_table("unit_version", schema=None) as batch_op:
        batch_op.alter_column(
            "name", existing_type=sa.VARCHAR(length=255), nullable=False
        )


def downgrade():
    with op.batch_alter_table("unit_version", schema=None) as batch_op:
        batch_op.alter_column(
            "name", existing_type=sa.VARCHAR(length=255), nullable=True
        )

    with op.batch_alter_table("transaction", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "remote_addr", sa.VARCHAR(length=50), autoincrement=False, nullable=True
            )
        )
        batch_op.alter_column("issued_at", existing_type=sa.DateTime(), nullable=True)

    with op.batch_alter_table("scenario_version", schema=None) as batch_op:
        batch_op.alter_column(
            "name", existing_type=sa.VARCHAR(length=255), nullable=True
        )

    with op.batch_alter_table("runmetaentry_version", schema=None) as batch_op:
        batch_op.alter_column(
            "dtype", existing_type=sa.VARCHAR(length=20), nullable=True
        )
        batch_op.alter_column(
            "key", existing_type=sa.VARCHAR(length=1023), nullable=True
        )
        batch_op.alter_column("run__id", existing_type=sa.INTEGER(), nullable=True)

    with op.batch_alter_table("run_version", schema=None) as batch_op:
        batch_op.alter_column("is_default", existing_type=sa.BOOLEAN(), nullable=True)
        batch_op.alter_column("version", existing_type=sa.INTEGER(), nullable=True)
        batch_op.alter_column("scenario__id", existing_type=sa.INTEGER(), nullable=True)
        batch_op.alter_column("model__id", existing_type=sa.INTEGER(), nullable=True)

    with op.batch_alter_table("region_version", schema=None) as batch_op:
        batch_op.alter_column(
            "hierarchy", existing_type=sa.VARCHAR(length=1023), nullable=True
        )
        batch_op.alter_column(
            "name", existing_type=sa.VARCHAR(length=255), nullable=True
        )

    with op.batch_alter_table("model_version", schema=None) as batch_op:
        batch_op.alter_column(
            "name", existing_type=sa.VARCHAR(length=255), nullable=True
        )

    with op.batch_alter_table("iamc_variable_version", schema=None) as batch_op:
        batch_op.alter_column(
            "name", existing_type=sa.VARCHAR(length=255), nullable=True
        )

    with op.batch_alter_table("iamc_timeseries_version", schema=None) as batch_op:
        batch_op.alter_column("run__id", existing_type=sa.INTEGER(), nullable=True)
        batch_op.alter_column(
            "measurand__id", existing_type=sa.INTEGER(), nullable=True
        )
        batch_op.alter_column("region__id", existing_type=sa.INTEGER(), nullable=True)

    with op.batch_alter_table("iamc_measurand_version", schema=None) as batch_op:
        batch_op.alter_column("unit__id", existing_type=sa.INTEGER(), nullable=True)
        batch_op.alter_column("variable__id", existing_type=sa.INTEGER(), nullable=True)

    with op.batch_alter_table(
        "iamc_datapoint_universal_version", schema=None
    ) as batch_op:
        batch_op.alter_column(
            "type", existing_type=sa.VARCHAR(length=255), nullable=True
        )
        batch_op.alter_column(
            "time_series__id", existing_type=sa.INTEGER(), nullable=True
        )

    with op.batch_alter_table("checkpoint", schema=None) as batch_op:
        batch_op.alter_column(
            "transaction__id", existing_type=sa.INTEGER(), nullable=False
        )



from typing import ClassVar

from ixmp4 import db
from ixmp4.data import types
from ixmp4.data.abstract import iamc as abstract
from ixmp4.data.db import mixins, versions

from .. import base


class Variable(base.BaseModel, mixins.HasCreationInfo):
    __tablename__ = "iamc_variable"

    NotFound: ClassVar = abstract.Variable.NotFound
    NotUnique: ClassVar = abstract.Variable.NotUnique
    DeletionPrevented: ClassVar = abstract.Variable.DeletionPrevented

    name: types.UniqueName


class VariableVersion(versions.DefaultVersionModel):
    __tablename__ = "iamc_variable_version"
    name: types.String = db.Column(db.String(255), nullable=False)

    created_at: types.DateTime = db.Column(nullable=True)
    created_by: types.Username


version_triggers = versions.PostgresVersionTriggers(
    Variable.__table__, VariableVersion.__table__
)



from typing import TYPE_CHECKING

import pandas as pd

from ixmp4.data.abstract.checkpoint import Checkpoint
from ixmp4.data.backend import Backend

from .base import BaseFacade

if TYPE_CHECKING:
    from .run import Run


class RunCheckpoints(BaseFacade):
    def __init__(self, run: "Run", _backend: Backend | None = None) -> None:
        self.run = run
        super().__init__(_backend)

    def tabulate(self) -> pd.DataFrame:
        return self.backend.checkpoints.tabulate(run__id=self.run.id)

    def create(self, message: str) -> Checkpoint:
        self.run.require_lock()

        return self.backend.checkpoints.create(run__id=self.run.id, message=message)



from .base import BaseModel, DataFrame
from .checkpoint import Checkpoint, CheckpointRepository
from .docs import Docs, DocsRepository
from .iamc import (  # Measurand,; MeasurandRepository,; AnnualDataPoint,; SubAnnualDataPoint,; CategoricalDataPoint, # noqa: E501
    DataPoint,
    DataPointRepository,
    TimeSeries,
    TimeSeriesRepository,
    Variable,
    VariableRepository,
)
from .meta import RunMetaEntry, RunMetaEntryRepository
from .model import Model, ModelRepository
from .optimization import (
    Equation,
    EquationRepository,
    IndexSet,
    IndexSetRepository,
    Parameter,
    ParameterRepository,
    Scalar,
    ScalarRepository,
    Table,
    TableRepository,
)

# TODO for PR: avoiding name conflict here Is that okay?
from .optimization import Variable as OptimizationVariable
from .optimization import VariableRepository as OptimizationVariableRepository
from .region import Region, RegionParent, RegionRepository
from .run import Run, RunRepository
from .scenario import Scenario, ScenarioRepository
from .unit import Unit, UnitRepository



# type: ignore
"""Remove Column class for Parameter

Revision ID: 86fd6ecaae8a
Revises: bc2e195788ba
Create Date: 2025-04-28 11:29:06.675241

"""

import sqlalchemy as sa
from alembic import op

# Revision identifiers, used by Alembic.
revision = "86fd6ecaae8a"
down_revision = "bc2e195788ba"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "optimization_parameterindexsetassociation",
        sa.Column("parameter__id", sa.Integer(), nullable=False),
        sa.Column("indexset__id", sa.Integer(), nullable=False),
        sa.Column("column_name", sa.String(length=255), nullable=True),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["indexset__id"],
            ["optimization_indexset.id"],
            name=op.f("fk_optimization_parameterindexsetassociation_indexset___7045"),
        ),
        sa.ForeignKeyConstraint(
            ["parameter__id"],
            ["optimization_parameter.id"],
            name=op.f("fk_optimization_parameterindexsetassociation_parameter__8a6d"),
        ),
        sa.PrimaryKeyConstraint(
            "id", name=op.f("pk_optimization_parameterindexsetassociation")
        ),
    )
    with op.batch_alter_table(
        "optimization_parameterindexsetassociation", schema=None
    ) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_optimization_parameterindexsetassociation_indexset__id"),
            ["indexset__id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_optimization_parameterindexsetassociation_parameter__id"),
            ["parameter__id"],
            unique=False,
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table(
        "optimization_parameterindexsetassociation", schema=None
    ) as batch_op:
        batch_op.drop_index(
            batch_op.f("ix_optimization_parameterindexsetassociation_parameter__id")
        )
        batch_op.drop_index(
            batch_op.f("ix_optimization_parameterindexsetassociation_indexset__id")
        )

    op.drop_table("optimization_parameterindexsetassociation")
    # ### end Alembic commands ###



# type: ignore
"""Remove Column class entirely

Revision ID: 8fd96526b04f
Revises: 86fd6ecaae8a
Create Date: 2025-04-29 13:23:48.513452

"""

import sqlalchemy as sa
from alembic import op

# Revision identifiers, used by Alembic.
revision = "8fd96526b04f"
down_revision = "86fd6ecaae8a"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "optimization_equationindexsetassociation",
        sa.Column("equation__id", sa.Integer(), nullable=False),
        sa.Column("indexset__id", sa.Integer(), nullable=False),
        sa.Column("column_name", sa.String(length=255), nullable=True),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["equation__id"],
            ["optimization_equation.id"],
            name=op.f("fk_optimization_equationindexsetassociation_equation__i_be8b"),
        ),
        sa.ForeignKeyConstraint(
            ["indexset__id"],
            ["optimization_indexset.id"],
            name=op.f("fk_optimization_equationindexsetassociation_indexset__i_82d0"),
        ),
        sa.PrimaryKeyConstraint(
            "id", name=op.f("pk_optimization_equationindexsetassociation")
        ),
    )
    with op.batch_alter_table(
        "optimization_equationindexsetassociation", schema=None
    ) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_optimization_equationindexsetassociation_equation__id"),
            ["equation__id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_optimization_equationindexsetassociation_indexset__id"),
            ["indexset__id"],
            unique=False,
        )

    op.create_table(
        "optimization_tableindexsetassociation",
        sa.Column("table__id", sa.Integer(), nullable=False),
        sa.Column("indexset__id", sa.Integer(), nullable=False),
        sa.Column("column_name", sa.String(length=255), nullable=True),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["indexset__id"],
            ["optimization_indexset.id"],
            name=op.f("fk_optimization_tableindexsetassociation_indexset__id_o_868a"),
        ),
        sa.ForeignKeyConstraint(
            ["table__id"],
            ["optimization_table.id"],
            name=op.f("fk_optimization_tableindexsetassociation_table__id_opti_5cfe"),
        ),
        sa.PrimaryKeyConstraint(
            "id", name=op.f("pk_optimization_tableindexsetassociation")
        ),
    )
    with op.batch_alter_table(
        "optimization_tableindexsetassociation", schema=None
    ) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_optimization_tableindexsetassociation_indexset__id"),
            ["indexset__id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_optimization_tableindexsetassociation_table__id"),
            ["table__id"],
            unique=False,
        )

    op.create_table(
        "optimization_variableindexsetassociation",
        sa.Column("variable__id", sa.Integer(), nullable=False),
        sa.Column("indexset__id", sa.Integer(), nullable=False),
        sa.Column("column_name", sa.String(length=255), nullable=True),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["indexset__id"],
            ["optimization_indexset.id"],
            name=op.f("fk_optimization_variableindexsetassociation_indexset__i_044c"),
        ),
        sa.ForeignKeyConstraint(
            ["variable__id"],
            ["optimization_optimizationvariable.id"],
            name=op.f("fk_optimization_variableindexsetassociation_variable__i_02a8"),
        ),
        sa.PrimaryKeyConstraint(
            "id", name=op.f("pk_optimization_variableindexsetassociation")
        ),
    )
    with op.batch_alter_table(
        "optimization_variableindexsetassociation", schema=None
    ) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_optimization_variableindexsetassociation_indexset__id"),
            ["indexset__id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_optimization_variableindexsetassociation_variable__id"),
            ["variable__id"],
            unique=False,
        )

    op.drop_table("optimization_column_docs")
    with op.batch_alter_table("optimization_column", schema=None) as batch_op:
        batch_op.drop_index("ix_optimization_column_constrained_to_indexset")

    op.drop_table("optimization_column")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "optimization_column",
        sa.Column("name", sa.VARCHAR(length=255), nullable=False),
        sa.Column("dtype", sa.VARCHAR(length=255), nullable=False),
        sa.Column("table__id", sa.INTEGER(), nullable=True),
        sa.Column("constrained_to_indexset", sa.INTEGER(), nullable=False),
        sa.Column("unique", sa.BOOLEAN(), nullable=False),
        sa.Column("id", sa.INTEGER(), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("created_by", sa.VARCHAR(length=255), nullable=True),
        sa.Column("equation__id", sa.INTEGER(), nullable=True),
        sa.Column("parameter__id", sa.INTEGER(), nullable=True),
        sa.Column("variable__id", sa.INTEGER(), nullable=True),
        sa.ForeignKeyConstraint(
            ["constrained_to_indexset"],
            ["optimization_indexset.id"],
            name="fk_optimization_column_constrained_to_indexset_optimiza_8432",
        ),
        sa.ForeignKeyConstraint(
            ["equation__id"],
            ["optimization_equation.id"],
            name="fk_optimization_column_equation__id_optimization_equation",
        ),
        sa.ForeignKeyConstraint(
            ["parameter__id"],
            ["optimization_parameter.id"],
            name="fk_optimization_column_parameter__id_optimization_parameter",
        ),
        sa.ForeignKeyConstraint(
            ["table__id"],
            ["optimization_table.id"],
            name="fk_optimization_column_table__id_optimization_table",
        ),
        sa.ForeignKeyConstraint(
            ["variable__id"],
            ["optimization_optimizationvariable.id"],
            name="fk_optimization_column_variable__id_optimization_optimi_2f05",
        ),
        sa.PrimaryKeyConstraint("id", name="pk_optimization_column"),
        sa.UniqueConstraint(
            "name", "table__id", name="uq_optimization_column_name_table__id"
        ),
    )
    with op.batch_alter_table("optimization_column", schema=None) as batch_op:
        batch_op.create_index(
            "ix_optimization_column_constrained_to_indexset",
            ["constrained_to_indexset"],
            unique=False,
        )

    op.create_table(
        "optimization_column_docs",
        sa.Column("description", sa.TEXT(), nullable=False),
        sa.Column("dimension__id", sa.INTEGER(), nullable=True),
        sa.Column("id", sa.INTEGER(), nullable=False),
        sa.ForeignKeyConstraint(
            ["dimension__id"],
            ["optimization_column.id"],
            name="fk_optimization_column_docs_dimension__id_optimization_column",
        ),
        sa.PrimaryKeyConstraint("id", name="pk_optimization_column_docs"),
        sa.UniqueConstraint(
            "dimension__id", name="uq_optimization_column_docs_dimension__id"
        ),
    )
    with op.batch_alter_table(
        "optimization_variableindexsetassociation", schema=None
    ) as batch_op:
        batch_op.drop_index(
            batch_op.f("ix_optimization_variableindexsetassociation_variable__id")
        )
        batch_op.drop_index(
            batch_op.f("ix_optimization_variableindexsetassociation_indexset__id")
        )

    op.drop_table("optimization_variableindexsetassociation")
    with op.batch_alter_table(
        "optimization_tableindexsetassociation", schema=None
    ) as batch_op:
        batch_op.drop_index(
            batch_op.f("ix_optimization_tableindexsetassociation_table__id")
        )
        batch_op.drop_index(
            batch_op.f("ix_optimization_tableindexsetassociation_indexset__id")
        )

    op.drop_table("optimization_tableindexsetassociation")
    with op.batch_alter_table(
        "optimization_equationindexsetassociation", schema=None
    ) as batch_op:
        batch_op.drop_index(
            batch_op.f("ix_optimization_equationindexsetassociation_indexset__id")
        )
        batch_op.drop_index(
            batch_op.f("ix_optimization_equationindexsetassociation_equation__id")
        )

    op.drop_table("optimization_equationindexsetassociation")
    # ### end Alembic commands ###



from typing import Any, ClassVar

from ixmp4 import db
from ixmp4.data.db import filters as base
from ixmp4.data.db.iamc.timeseries import TimeSeries
from ixmp4.data.db.run.model import Run
from ixmp4.db import filters, typing_column, utils


class IamcRunFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    region: base.RegionFilter
    variable: base.VariableFilter
    unit: base.UnitFilter

    sqla_model: ClassVar[type] = Run

    _remote_filters = {"region", "variable", "unit"}
    _remote_path = [
        {
            "target_model": TimeSeries,
            "fk_attr": "run__id",
            "source_model": Run,
            "pk_attr": "id",
        },
    ]

    def join(
        self, exc: db.sql.Select[tuple[Run]], session: db.Session | None = None
    ) -> db.sql.Select[tuple[Run]]:
        if self._should_use_subquery_optimization():
            return exc

        if not utils.is_joined(exc, TimeSeries):
            exc = exc.join(TimeSeries, onclause=TimeSeries.run__id == Run.id)
        return exc


class RunFilter(base.RunFilter, metaclass=filters.FilterMeta):
    iamc: IamcRunFilter | filters.Boolean | None = None

    def join_datapoints(
        self, exc: db.sql.Select[tuple[Run]], session: db.Session | None = None
    ) -> db.sql.Select[tuple[Run]]:
        if not utils.is_joined(exc, TimeSeries):
            exc = exc.join(TimeSeries, onclause=TimeSeries.run__id == Run.id)
        return exc

    def filter_iamc(
        self,
        exc: db.sql.Select[tuple[Run]],
        c: typing_column[Any],  # Any since it is unused
        v: bool | None,
        session: db.Session | None = None,
    ) -> db.sql.Select[tuple[Run]]:
        if v is None:
            return exc

        if v is True:
            return self.join_datapoints(exc, session)
        else:
            ids = self.join_datapoints(db.select(Run.id), session)
            exc = exc.where(~Run.id.in_(ids))
            return exc

    def join(
        self, exc: db.sql.Select[tuple[Run]], session: db.Session | None = None
    ) -> db.sql.Select[tuple[Run]]:
        return exc



# type: ignore
"""load initial version data

Revision ID: 84b534bbc858
Revises: 1e5554a164e6
Create Date: 2025-05-28 17:25:06.159790

"""

import enum
import logging
from datetime import datetime

import sqlalchemy as sa
from alembic import op

revision = "84b534bbc858"
down_revision = "1e5554a164e6"
branch_labels = None
depends_on = None


class Operation(int, enum.Enum):
    INSERT = 0
    UPDATE = 1
    DELETE = 2


logger = logging.getLogger(__name__)

initial_transaction_id = 1
tables = [
    ("iamc_datapoint_universal", "iamc_datapoint_universal_version"),
    ("iamc_measurand", "iamc_measurand_version"),
    ("iamc_timeseries", "iamc_timeseries_version"),
    ("iamc_variable", "iamc_variable_version"),
    ("model", "model_version"),
    ("region", "region_version"),
    ("run", "run_version"),
    ("runmetaentry", "runmetaentry_version"),
    ("scenario", "scenario_version"),
    ("unit", "unit_version"),
]

tabledata_chunksize = 10000


def generate_initial_version_data(
    data_table_name: str, version_table_name: str, transaction_id: int
):
    """Generate initial version data for a given data table."""
    logging.info(
        f"Copying initial data from '{data_table_name}' "
        f"to '{version_table_name}' table ..."
    )

    conn = op.get_bind()

    if conn is not None:
        version_table = sa.Table(
            version_table_name,
            sa.MetaData(),
            autoload_with=conn,
        )

        data_table = sa.Table(
            data_table_name,
            sa.MetaData(),
            autoload_with=conn,
        )

        partitions = (
            conn.execute(
                sa.select(data_table).execution_options(yield_per=tabledata_chunksize)
            )
            .mappings()
            .partitions(tabledata_chunksize)
        )
        for res in partitions:
            initial_versions = [
                {
                    **row,
                    "transaction_id": transaction_id,
                    "end_transaction_id": None,
                    "operation_type": Operation.INSERT,
                }
                for row in res
            ]
            op.bulk_insert(version_table, initial_versions)
    else:
        logging.warning("No database connection available.")


def upgrade():
    logging.info(
        f"Creating initial transaction object with ID {initial_transaction_id} ..."
    )

    conn = op.get_bind()

    if conn is not None:
        if conn.dialect.name != "postgresql":
            logging.info("Not running on postgres database, skipping migration.")
            return

        # reflect transaction table
        transaction_table = sa.Table(
            "transaction",
            sa.MetaData(),
            autoload_with=conn,
        )

        # Check if any transactions already exist
        res = conn.execute(sa.select(transaction_table)).all()
        if len(res) > 0:
            logging.warning(
                "Database already contains transactions. "
                "Skipping creation of initial version "
                "data to avoid corrupting the database."
            )
            return

        # Create initial transaction
        op.bulk_insert(
            transaction_table,
            [
                {
                    "id": initial_transaction_id,
                    "remote_addr": "",
                    "issued_at": datetime.now(),
                },
            ],
        )

    else:
        logging.warning("No database connection available.")

    for data_table_name, version_table_name in tables:
        generate_initial_version_data(
            data_table_name,
            version_table_name,
            initial_transaction_id,
        )


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###



from typing import Any

from ixmp4 import db
from ixmp4.data.db import filters as base
from ixmp4.data.db.iamc.measurand import Measurand
from ixmp4.data.db.iamc.timeseries import TimeSeries
from ixmp4.db import filters, typing_column, utils

from . import Unit


class BaseIamcFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    def join_datapoints(
        self, exc: db.sql.Select[tuple[Unit]], session: db.Session | None = None
    ) -> db.sql.Select[tuple[Unit]]:
        if not utils.is_joined(exc, Measurand):
            exc = exc.join(Measurand, Measurand.unit__id == Unit.id)

        if not utils.is_joined(exc, TimeSeries):
            exc = exc.join(
                TimeSeries, onclause=TimeSeries.measurand__id == Measurand.id
            )
        return exc


class SimpleIamcUnitFilter(
    base.UnitFilter, BaseIamcFilter, metaclass=filters.FilterMeta
):
    def join(
        self, exc: db.sql.Select[tuple[Unit]], session: db.Session | None = None
    ) -> db.sql.Select[tuple[Unit]]:
        return super().join_datapoints(exc, session)


class IamcUnitFilter(
    base.UnitFilter,
    BaseIamcFilter,
    metaclass=filters.FilterMeta,
):
    variable: base.VariableFilter | None = filters.Field(None)
    region: base.RegionFilter | None = filters.Field(None)
    run: base.RunFilter = filters.Field(
        default=base.RunFilter(id=None, version=None, is_default=True)
    )

    _remote_filters = {"run", "region", "variable"}
    _remote_path = [
        {
            "target_model": Measurand,
            "fk_attr": "unit__id",
            "source_model": Unit,
            "pk_attr": "id",
        },
        {
            "target_model": TimeSeries,
            "fk_attr": "measurand__id",
            "source_model": Measurand,
            "pk_attr": "id",
        },
    ]

    def join(
        self, exc: db.sql.Select[tuple[Unit]], session: db.Session | None = None
    ) -> db.sql.Select[tuple[Unit]]:
        if self._should_use_subquery_optimization():
            return exc

        return super().join_datapoints(exc, session)


class UnitFilter(base.UnitFilter, BaseIamcFilter, metaclass=filters.FilterMeta):
    iamc: IamcUnitFilter | filters.Boolean

    def filter_iamc(
        self,
        exc: db.sql.Select[tuple[Unit]],
        c: typing_column[Any],  # Any since it is unused
        v: bool | None,
        session: db.Session | None = None,
    ) -> db.sql.Select[tuple[Unit]]:
        if v is None:
            return exc

        if v is True:
            return self.join_datapoints(exc, session)
        else:
            ids = self.join_datapoints(db.select(Unit.id), session)
            exc = exc.where(~Unit.id.in_(ids))
            return exc

    def join(
        self, exc: db.sql.Select[tuple[Unit]], session: db.Session | None = None
    ) -> db.sql.Select[tuple[Unit]]:
        return exc



from collections.abc import Iterable
from typing import TYPE_CHECKING

import pandas as pd

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import Unpack

from ixmp4.data.abstract import DataPoint as DataPointModel
from ixmp4.data.abstract.iamc.datapoint import EnumerateKwargs
from ixmp4.data.backend import Backend
from ixmp4.data.db.iamc.utils import (
    AddDataPointFrameSchema,
    RemoveDataPointFrameSchema,
    normalize_df,
)

from ..base import BaseFacade
from ..utils import substitute_type
from .variable import VariableRepository

if TYPE_CHECKING:
    from ..run import Run


class RunIamcData(BaseFacade):
    """IAMC data.

    Parameters
    ----------
    backend : ixmp4.data.backend.Backend
        Data source backend.
    run : ixmp4.base.run.Run
        Model run.
    """

    run: "Run"

    def __init__(self, run: "Run", **kwargs: Backend | None) -> None:
        super().__init__(**kwargs)
        self.run = run

    def _get_or_create_ts(self, df: pd.DataFrame) -> pd.DataFrame:
        id_cols = ["region", "variable", "unit", "run__id"]
        # create set of unqiue timeseries (if missing)
        ts_df = df[id_cols].drop_duplicates()
        self.backend.iamc.timeseries.bulk_upsert(ts_df, create_related=True)

        # retrieve them again to get database ids
        ts_df = self.backend.iamc.timeseries.tabulate(
            join_parameters=True,
            run={"id": self.run.id, "default_only": False},
        )
        ts_df = ts_df.rename(columns={"id": "time_series__id"})

        # merge on the identity columns
        return pd.merge(
            df, ts_df, how="left", on=id_cols, suffixes=(None, "_y")
        )  # tada, df with 'time_series__id' added from the database.

    def add(self, df: pd.DataFrame, type: DataPointModel.Type | None = None) -> None:
        self.run.require_lock()
        df = AddDataPointFrameSchema.validate(df)
        df["run__id"] = self.run.id
        df = self._get_or_create_ts(df)
        substitute_type(df, type)
        self.backend.iamc.datapoints.bulk_upsert(df)

    def remove(self, df: pd.DataFrame, type: DataPointModel.Type | None = None) -> None:
        self.run.require_lock()
        df = RemoveDataPointFrameSchema.validate(df)
        df["run__id"] = self.run.id
        df = self._get_or_create_ts(df)
        substitute_type(df, type)
        df = df.drop(columns=["unit", "variable", "region"])
        self.backend.iamc.datapoints.bulk_delete(df)

    def tabulate(
        self,
        *,
        variable: dict[str, str | Iterable[str]] | None = None,
        region: dict[str, str | Iterable[str]] | None = None,
        unit: dict[str, str | Iterable[str]] | None = None,
        raw: bool = False,
    ) -> pd.DataFrame:
        df = self.backend.iamc.datapoints.tabulate(
            join_parameters=True,
            join_runs=False,
            run={"id": self.run.id, "default_only": False},
            variable=variable,
            region=region,
            unit=unit,
        ).dropna(how="all", axis="columns")
        return normalize_df(df, raw, False, False)


class PlatformIamcData(BaseFacade):
    variables: VariableRepository

    def __init__(self, _backend: Backend | None = None) -> None:
        self.variables = VariableRepository(_backend=_backend)
        super().__init__(_backend=_backend)

    def tabulate(
        self,
        *,
        join_runs: bool = True,
        join_run_id: bool = False,
        raw: bool = False,
        **kwargs: Unpack[EnumerateKwargs],
    ) -> pd.DataFrame:
        df = self.backend.iamc.datapoints.tabulate(
            join_parameters=True,
            join_runs=join_runs,
            join_run_id=join_run_id,
            **kwargs,
        ).dropna(how="all", axis="columns")
        return normalize_df(df, raw, join_runs, join_run_id)



from typing import TypeVar

import pandas as pd
import pandera as pa
from pandera.engines import pandas_engine
from pandera.pandas import DataFrameModel
from pandera.typing import Series

from ixmp4.data.abstract import DataPoint as DataPointModel


class RemoveDataPointFrameSchema(DataFrameModel):
    type: Series[pa.String] | None = pa.Field(isin=[t for t in DataPointModel.Type])
    step_year: Series[pa.Int] | None = pa.Field(coerce=True, nullable=True)
    step_datetime: Series[pandas_engine.DateTime] | None = pa.Field(
        coerce=True, nullable=True
    )
    step_category: Series[pa.String] | None = pa.Field(nullable=True)

    region: Series[pa.String] | None = pa.Field(coerce=True)
    unit: Series[pa.String] | None = pa.Field(coerce=True)
    variable: Series[pa.String] | None = pa.Field(coerce=True)


class AddDataPointFrameSchema(RemoveDataPointFrameSchema):
    value: Series[pa.Float] = pa.Field(coerce=True)


MAP_STEP_COLUMN = {
    "ANNUAL": "step_year",
    "CATEGORICAL": "step_year",
    "DATETIME": "step_datetime",
}


def convert_to_std_format(
    df: pd.DataFrame, join_runs: bool, join_run_id: bool
) -> pd.DataFrame:
    df.rename(columns={"step_category": "subannual"}, inplace=True)

    if set(df.type.unique()).issubset(["ANNUAL", "CATEGORICAL"]):
        df.rename(columns={"step_year": "year"}, inplace=True)
        time_col = "year"
    else:
        T = TypeVar("T", bool, float, int, str)

        def map_step_column(df: "pd.Series[T]") -> "pd.Series[T]":
            df["time"] = df[MAP_STEP_COLUMN[str(df.type)]]
            return df

        df = df.apply(map_step_column, axis=1)
        time_col = "time"

    columns = []
    if join_run_id and "run__id" in df.columns:
        columns.append("run__id")
    if join_runs:
        columns.extend(["model", "scenario", "version"])
    columns += ["region", "variable", "unit"] + [time_col]
    if "subannual" in df.columns:
        columns += ["subannual"]
    return df[columns + ["value"]]


def normalize_df(
    df: pd.DataFrame, raw: bool, join_runs: bool, join_run_id: bool
) -> pd.DataFrame:
    if not df.empty:
        df = df.drop(columns=["time_series__id"])
        if raw is False:
            return convert_to_std_format(df, join_runs, join_run_id)
    return df



import time
import warnings
from collections import UserDict
from contextlib import contextmanager
from typing import ClassVar, Generator, cast

import numpy as np
import pandas as pd

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import TypedDict, Unpack

from ixmp4.core.exceptions import OperationNotSupported
from ixmp4.data.abstract import Model as ModelModel
from ixmp4.data.abstract import Run as RunModel
from ixmp4.data.abstract import Scenario as ScenarioModel
from ixmp4.data.abstract.annotations import PrimitiveTypes
from ixmp4.data.abstract.run import EnumerateKwargs
from ixmp4.data.backend import Backend

from .base import BaseFacade, BaseModelFacade
from .checkpoints import RunCheckpoints
from .exceptions import RunLockRequired
from .iamc import RunIamcData
from .optimization import OptimizationData


class RunKwargs(TypedDict):
    _backend: Backend
    _model: RunModel


class Run(BaseModelFacade):
    _model: RunModel
    _meta: "RunMetaFacade"
    NoDefaultVersion: ClassVar = RunModel.NoDefaultVersion
    NotFound: ClassVar = RunModel.NotFound
    NotUnique: ClassVar = RunModel.NotUnique

    checkpoints: RunCheckpoints

    owns_lock: bool = False
    minimum_lock_timeout: float = 0.1
    maximum_lock_timeout: float = 5

    def __init__(self, **kwargs: Unpack[RunKwargs]) -> None:
        super().__init__(**kwargs)

        self.version = self._model.version

        self.iamc = RunIamcData(_backend=self.backend, run=self)
        self._meta = RunMetaFacade(_backend=self.backend, run=self)
        self.optimization = OptimizationData(_backend=self.backend, run=self)
        self.checkpoints = RunCheckpoints(_backend=self.backend, run=self)

    @property
    def model(self) -> ModelModel:
        """Associated model."""
        return self._model.model

    @property
    def scenario(self) -> ScenarioModel:
        """Associated scenario."""
        return self._model.scenario

    @property
    def id(self) -> int:
        """Unique id."""
        return self._model.id

    @property
    def meta(self) -> "RunMetaFacade":
        "Meta indicator data (`dict`-like)."
        return self._meta

    @meta.setter
    def meta(self, meta: dict[str, PrimitiveTypes | np.generic | None]) -> None:
        self._meta._set(meta)

    @property
    def is_default(self) -> bool:
        return self._model.is_default

    def set_as_default(self) -> None:
        """Sets this run as the default version for its `model` + `scenario`
        combination."""
        self.backend.runs.set_as_default_version(self._model.id)
        self._model = self.backend.runs.get_by_id(self._model.id)

    def unset_as_default(self) -> None:
        """Unsets this run as the default version."""
        self.backend.runs.unset_as_default_version(self._model.id)
        self._model = self.backend.runs.get_by_id(self._model.id)

    def require_lock(self) -> None:
        if not self.owns_lock:
            raise RunLockRequired()

    def _lock(self) -> None:
        self._model = self.backend.runs.lock(self._model.id)
        self.owns_lock = True

    def _unlock(self) -> None:
        self._model = self.backend.runs.unlock(self._model.id)
        self.owns_lock = False

    def _lock_with_timeout(self, timeout: float) -> None:
        """Try locking the run until a timeout passes."""
        start_time = time.time()
        while True:
            try:
                self._lock()
                break
            except RunModel.IsLocked as e:
                elapsed_time = time.time() - start_time
                if elapsed_time > timeout:
                    raise e
                remaining_time = timeout - elapsed_time
                sleep_time = elapsed_time * 2
                sleep_time = min(sleep_time, self.maximum_lock_timeout)
                sleep_time = max(sleep_time, self.minimum_lock_timeout)
                sleep_time = min(sleep_time, remaining_time)
                time.sleep(sleep_time)

    @contextmanager
    def transact(
        self,
        message: str,
        timeout: float | None = None,
        revert_platform_on_error: bool = False,
    ) -> Generator[None, None, None]:
        """
        Context manager to lock the run before yielding control
        back to the caller. The run is unlocked and a checkpoint
        with the provided `message` is created after the context
        manager exits.
        If an exception occurs, the run is reverted to the last
        checkpoint or if no checkpoint exists, to the transaction
        the run was locked at.

        If the run is already locked, the context manager will
        throw `Run.IsLocked` or if `timeout` is provided retry until
        the timeout has passed (and then throw the original
        `Run.IsLocked` exception).

        Parameters
        ----------
        messsage : str
            The message for the checkpoint created after
            conclusion of the context manager.
        timeout : int, optional
            Timeout in seconds.
        revert_platform_on_error : bool, optional
            Whether to revert units when encountering an error; default `False`.

        Raises
        ------
        :class:`ixmp4.core.exceptions.RunIsLocked`
            If the run is already locked and no timeout is provided
            or the provided timeout is exceeded.
        """

        if timeout is None:
            self._lock()
        else:
            self._lock_with_timeout(timeout)

        try:
            yield
        except Exception as e:
            checkpoint_df = self.checkpoints.tabulate()
            if checkpoint_df.empty:
                checkpoint_transaction = -1
            else:
                max_tx_id = checkpoint_df["transaction__id"].max()
                if pd.isnull(max_tx_id):
                    checkpoint_transaction = -1
                else:
                    checkpoint_transaction = int(max_tx_id)

            assert self._model.lock_transaction is not None

            target_transaction = max(
                checkpoint_transaction, self._model.lock_transaction
            )
            try:
                self.backend.runs.revert(
                    self._model.id,
                    target_transaction,
                    revert_platform=revert_platform_on_error,
                )
            except OperationNotSupported as ons_exc:
                warnings.warn(
                    "An exception occurred but the `Run` "
                    "was not reverted because versioning "
                    "is not supported by this platform: " + ons_exc.message
                )

            self._meta.refetch_data()
            self._unlock()
            raise e

        self.checkpoints.create(message)
        self._unlock()

    def delete(self) -> None:
        """Delete this run.
        Tries to acquire a lock in the background.

        Raises
        ------
        :class:`ixmp4.core.exceptions.RunIsLocked`:
            If the run is already locked by this or another object.
        """
        self.backend.runs.delete(self._model.id)

    def clone(
        self,
        model: str | None = None,
        scenario: str | None = None,
        keep_solution: bool = True,
    ) -> "Run":
        return Run(
            _backend=self.backend,
            _model=self.backend.runs.clone(
                run_id=self.id,
                model_name=model,
                scenario_name=scenario,
                keep_solution=keep_solution,
            ),
        )


class RunRepository(BaseFacade):
    def create(self, model: str, scenario: str) -> Run:
        return Run(
            _backend=self.backend, _model=self.backend.runs.create(model, scenario)
        )

    def delete(self, x: Run | int) -> None:
        if isinstance(x, Run):
            id = x.id
        elif isinstance(x, int):
            id = x
        else:
            raise TypeError("Invalid argument: Must be `Run` or `int`.")

        self.backend.runs.delete(id)

    def get(self, model: str, scenario: str, version: int | None = None) -> Run:
        _model = (
            self.backend.runs.get_default_version(model, scenario)
            if version is None
            else self.backend.runs.get(model, scenario, version)
        )
        return Run(_backend=self.backend, _model=_model)

    def list(self, **kwargs: Unpack[EnumerateKwargs]) -> list[Run]:
        return [
            Run(_backend=self.backend, _model=r)
            for r in self.backend.runs.list(**kwargs)
        ]

    def tabulate(
        self, audit_info: bool = False, **kwargs: Unpack[EnumerateKwargs]
    ) -> pd.DataFrame:
        runs = self.backend.runs.tabulate(**kwargs)
        runs["model"] = runs["model__id"].map(self.backend.models.map())
        runs["scenario"] = runs["scenario__id"].map(self.backend.scenarios.map())
        columns = ["model", "scenario", "version", "is_default"]
        if audit_info:
            columns += ["updated_at", "updated_by", "created_at", "created_by", "id"]
        return runs[columns]


class RunMetaFacade(BaseFacade, UserDict[str, PrimitiveTypes | None]):
    run: Run

    def __init__(self, run: Run, **kwargs: Backend) -> None:
        super().__init__(**kwargs)
        self.run = run
        self.refetch_data()

    def refetch_data(self) -> None:
        self.df, self.data = self._get()

    def _get(self) -> tuple[pd.DataFrame, dict[str, PrimitiveTypes | None]]:
        df = self.backend.meta.tabulate(run_id=self.run.id, run={"default_only": False})
        if df.empty:
            return df, {}
        return df, dict(zip(df["key"], df["value"]))

    def _set(self, meta: dict[str, PrimitiveTypes | np.generic | None]) -> None:
        self.run.require_lock()

        df = pd.DataFrame({"key": self.data.keys()})
        df["run__id"] = self.run.id
        self.backend.meta.bulk_delete(df)
        df = pd.DataFrame(
            {"key": meta.keys(), "value": [numpy_to_pytype(v) for v in meta.values()]}
        )
        df.dropna(axis=0, inplace=True)
        df["run__id"] = self.run.id
        self.backend.meta.bulk_upsert(df)
        self.df, self.data = self._get()

    def __setitem__(self, key: str, value: PrimitiveTypes | np.generic | None) -> None:
        self.run.require_lock()

        try:
            del self[key]
        except KeyError:
            pass

        py_value = numpy_to_pytype(value)
        if py_value is not None:
            self.backend.meta.create(self.run.id, key, py_value)
        self.df, self.data = self._get()

    def __delitem__(self, key: str) -> None:
        self.run.require_lock()
        id = dict(zip(self.df["key"], self.df["id"]))[key]
        self.backend.meta.delete(id)
        self.df, self.data = self._get()


def numpy_to_pytype(
    value: PrimitiveTypes | np.generic | None,
) -> PrimitiveTypes | None:
    """Cast numpy-types to basic Python types"""
    if value is np.nan:  # np.nan is cast to 'float', not None
        return None
    elif isinstance(value, np.generic):
        return cast(PrimitiveTypes, value.item())
    else:
        return value
